# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Optional, Tuple

import GPy
import numpy as np

from ..bayesian_optimization.interfaces import IEntropySearchModel
from ..core.interfaces import IDifferentiable, IJointlyDifferentiable, IModel, IModelWithNoise, IPriorHyperparameters
from ..experimental_design.interfaces import ICalculateVarianceReduction


class GPyModelWrapper(
    IModel,
    IDifferentiable,
    IJointlyDifferentiable,
    ICalculateVarianceReduction,
    IEntropySearchModel,
    IPriorHyperparameters,
    IModelWithNoise,
):
    """
    This is a thin wrapper around GPy models to allow users to plug GPy models into Emukit
    """

    def __init__(self, gpy_model: GPy.core.Model, n_restarts: int = 1):
        """
        :param gpy_model: GPy model object to wrap
        :param n_restarts: Number of restarts during hyper-parameter optimization
        """
        self.model = gpy_model
        self.n_restarts = n_restarts

    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param X: (n_points x n_dimensions) array containing locations at which to get predictions
        :return: (mean, variance) Arrays of size n_points x 1 of the predictive distribution at each input location
        """
        return self.model.predict(X)

    def predict_noiseless(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param X: (n_points x n_dimensions) array containing locations at which to get predictions
        :return: (mean, variance) Arrays of size n_points x 1 of the predictive distribution at each input location
        """
        return self.model.predict(X, include_likelihood=False)

    def predict_with_full_covariance(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param X: (n_points x n_dimensions) array containing locations at which to get predictions
        :return: (mean, variance) Arrays of size n_points x 1 and n_points x n_points of the predictive
                 mean and variance at each input location
        """
        return self.model.predict(X, full_cov=True)

    def get_prediction_gradients(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param X: (n_points x n_dimensions) array containing locations at which to get gradient of the predictions
        :return: (mean gradient, variance gradient) n_points x n_dimensions arrays of the gradients of the predictive
                 distribution at each input location
        """
        d_mean_dx, d_variance_dx = self.model.predictive_gradients(X)
        return d_mean_dx[:, :, 0], d_variance_dx

    def get_joint_prediction_gradients(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes and returns model gradients of mean and full covariance matrix at given points

        :param X: points to compute gradients at, nd array of shape (q, d)
        :return: Tuple with first item being gradient of the mean of shape (q) at X with respect to X (return shape is (q, q, d)).
                 The second item is the gradient of the full covariance matrix of shape (q, q) at X with respect to X
                 (return shape is (q, q, q, d)).
        """
        dmean_dx = dmean(X, self.model.X, self.model.kern, self.model.posterior.woodbury_vector[:, 0])
        dvariance_dx = dSigma(X, self.model.X, self.model.kern, self.model.posterior.woodbury_inv)
        return dmean_dx, dvariance_dx

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Sets training data in model

        :param X: New training features
        :param Y: New training outputs
        """
        self.model.set_XY(X, Y)

    def optimize(self, verbose=False):
        """
        Optimizes model hyper-parameters
        """
        self.model.optimize_restarts(self.n_restarts, verbose=verbose, robust=True)

    def calculate_variance_reduction(self, x_train_new: np.ndarray, x_test: np.ndarray) -> np.ndarray:
        """
        Computes the variance reduction at x_test, if a new point at x_train_new is acquired
        """
        covariance = self.model.posterior_covariance_between_points(x_train_new, x_test, include_likelihood=False)
        variance_prediction = self.model.predict(x_train_new)[1]
        return covariance**2 / variance_prediction

    def predict_covariance(self, X: np.ndarray, with_noise: bool = True) -> np.ndarray:
        """
        Calculates posterior covariance between points in X
        :param X: Array of size n_points x n_dimensions containing input locations to compute posterior covariance at
        :param with_noise: Whether to include likelihood noise in the covariance matrix
        :return: Posterior covariance matrix of size n_points x n_points
        """
        _, v = self.model.predict(X, full_cov=True, include_likelihood=with_noise)
        v = np.clip(v, 1e-10, np.inf)

        return v

    def get_covariance_between_points(self, X1: np.ndarray, X2: np.ndarray) -> np.ndarray:
        """
        Calculate posterior covariance between two sets of points.
        :param X1: An array of shape n_points1 x n_dimensions. This is the first argument of the
                   posterior covariance function.
        :param X2: An array of shape n_points2 x n_dimensions. This is the second argument of the
                   posterior covariance function.
        :return: An array of shape n_points1 x n_points2 of posterior covariances between X1 and X2.
            Namely, [i, j]-th entry of the returned array will represent the posterior covariance
            between i-th point in X1 and j-th point in X2.
        """
        return self.model.posterior_covariance_between_points(X1, X2, include_likelihood=False)

    @property
    def X(self) -> np.ndarray:
        """
        :return: An array of shape n_points x n_dimensions containing training inputs
        """
        return self.model.X

    @property
    def Y(self) -> np.ndarray:
        """
        :return: An array of shape n_points x 1 containing training outputs
        """
        return self.model.Y

    def generate_hyperparameters_samples(
        self, n_samples=20, n_burnin=100, subsample_interval=10, step_size=1e-1, leapfrog_steps=20
    ) -> np.ndarray:
        """
        Generates the samples from the hyper-parameters and returns them.
        :param n_samples: Number of generated samples.
        :param n_burnin: Number of initial samples not used.
        :param subsample_interval: Interval of subsampling from HMC samples.
        :param step_size: Size of the gradient steps in the HMC sampler.
        :param leapfrog_steps: Number of gradient steps before each Metropolis Hasting step.
        :return: A numpy array whose rows are samples of the hyper-parameters.

        """
        self.model.optimize(max_iters=self.n_restarts)
        # Add jitter to all unfixed parameters. After optimizing the hyperparameters, the gradient of the
        # posterior probability of the parameters wrt. the parameters will be close to 0.0, which is a poor
        # initialization for HMC
        unfixed_params = [param for param in self.model.flattened_parameters if not param.is_fixed]
        for param in unfixed_params:
            # Add jitter by multiplying with log-normal noise with mean 1 and standard deviation 0.01
            # This ensures the sign of the parameter remains the same
            param *= np.random.lognormal(np.log(1.0 / np.sqrt(1.0001)), np.sqrt(np.log(1.0001)), size=param.size)
        hmc = GPy.inference.mcmc.HMC(self.model, stepsize=step_size)
        samples = hmc.sample(num_samples=n_burnin + n_samples * subsample_interval, hmc_iters=leapfrog_steps)
        return samples[n_burnin::subsample_interval]

    def fix_model_hyperparameters(self, sample_hyperparameters: np.ndarray) -> None:
        """
        Fix model hyperparameters

        """
        if self.model._fixes_ is None:
            self.model[:] = sample_hyperparameters
        else:
            self.model[self.model._fixes_] = sample_hyperparameters
        self.model._trigger_params_changed()


def dSigma(x_predict: np.ndarray, x_train: np.ndarray, kern: GPy.kern, w_inv: np.ndarray) -> np.ndarray:
    """
    Compute the derivative of the posterior covariance with respect to the prediction input

    :param x_predict: Prediction inputs of shape (q, d)
    :param x_train: Training inputs of shape (n, d)
    :param kern: Covariance of the GP model
    :param w_inv: Woodbury inverse of the posterior fit of the GP
    :return: Gradient of the posterior covariance of shape (q, q, q, d)
    """
    q, d, n = x_predict.shape[0], x_predict.shape[1], x_train.shape[0]
    dkxX_dx = np.empty((q, n, d))
    dkxx_dx = np.empty((q, q, d))
    for i in range(d):
        dkxX_dx[:, :, i] = kern.dK_dX(x_predict, x_train, i)
        dkxx_dx[:, :, i] = kern.dK_dX(x_predict, x_predict, i)
    K = kern.K(x_predict, x_train)

    dsigma = np.zeros((q, q, q, d))
    for i in range(q):
        for j in range(d):
            Ks = np.zeros((q, n))
            Ks[i, :] = dkxX_dx[i, :, j]
            dKss_dxi = np.zeros((q, q))
            dKss_dxi[i, :] = dkxx_dx[i, :, j]
            dKss_dxi[:, i] = dkxx_dx[i, :, j].T
            dKss_dxi[i, i] = 0
            dsigma[:, :, i, j] = dKss_dxi - Ks @ w_inv @ K.T - K @ w_inv @ Ks.T
    return dsigma


def dmean(x_predict: np.ndarray, x_train: np.ndarray, kern: GPy.kern, w_vec: np.ndarray) -> np.ndarray:
    """
    Compute the derivative of the posterior mean with respect to prediction input

    :param x_predict: Prediction inputs of shape (q, d)
    :param x_train: Training inputs of shape (n, d)
    :param kern: Covariance of the GP model
    :param w_vec: Woodbury vector of the posterior fit of the GP
    :return: Gradient of the posterior mean of shape (q, q, d)
    """
    q, d, n = x_predict.shape[0], x_predict.shape[1], x_train.shape[0]
    dkxX_dx = np.empty((q, n, d))
    dmu = np.zeros((q, q, d))
    for i in range(d):
        dkxX_dx[:, :, i] = kern.dK_dX(x_predict, x_train, i)
        for j in range(q):
            dmu[j, j, i] = (dkxX_dx[j, :, i][None, :] @ w_vec[:, None]).flatten()
    return dmu


class GPyMultiOutputWrapper(IModel, IDifferentiable, ICalculateVarianceReduction, IEntropySearchModel):
    """
    A wrapper around GPy multi-output models.
    X inputs should have the corresponding output index as the last column in the array
    """

    def __init__(
        self, gpy_model: GPy.core.GP, n_outputs: int, n_optimization_restarts: int, verbose_optimization: bool = True
    ):
        """
        :param gpy_model: GPy multi-output model
        :param n_outputs: Number of outputs in the problem
        :param n_optimization_restarts: Number of restarts from random starting points when optimizing hyper-parameters
        """
        super().__init__()
        self.gpy_model = gpy_model
        self.n_optimization_restarts = n_optimization_restarts
        self.n_outputs = n_outputs
        self.verbose_optimization = verbose_optimization
        self.samples: Optional[np.ndarray] = None

    def calculate_variance_reduction(self, x_train_new: np.ndarray, x_test: np.ndarray) -> np.ndarray:
        """
        Calculates reduction in variance at x_test due to observing training point x_train_new

        :param x_train_new: New training point
        :param x_test: Test points to calculate variance reduction at
        :return: Array of variance reduction at each test point
        """
        fidelities_train_new = x_train_new[:, -1]
        y_metadata = {"output_index": fidelities_train_new.astype(int)}
        covariance = self.gpy_model.posterior_covariance_between_points(x_train_new, x_test, include_likelihood=False)
        variance_prediction = self.gpy_model.predict(x_train_new, Y_metadata=y_metadata)[1]
        return covariance**2 / variance_prediction

    def get_prediction_gradients(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculates gradients of predictions with respect to X, excluding with respect to the output index
        :param X: Point at which to predict gradients
        :return: (mean gradient, variance gradient)
        """
        dmean_dx, dvar_dx = self.gpy_model.predictive_gradients(X)
        return dmean_dx[:, :-1], dvar_dx[:, :-1]

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Predicts mean and variance for output specified by last column of X
        :param X: point(s) at which to predict
        :return: predicted (mean, variance) at X
        """
        output_index = X[:, -1]
        y_metadata = {"output_index": output_index.astype(int)}
        return self.gpy_model.predict(X, Y_metadata=y_metadata)

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Updates model with new training data
        :param X: New training features with output index as last column
        :param Y: New training targets with output index as last column
        """
        y_metadata = {"output_index": X[:, -1].astype(int)}
        self.gpy_model.update_model(False)
        self.gpy_model.Y_metadata = y_metadata
        self.gpy_model.set_XY(X, Y)
        self.gpy_model.update_model(True)

    def optimize(self) -> None:
        """
        Optimizes hyper-parameters of model. Starts the optimization at random locations equal to the values of the
        "n_optimization_restarts" attribute.
        """
        # Optimize the model if optimization_restarts > 0
        if self.n_optimization_restarts == 1:
            self.gpy_model.optimize()
        elif self.n_optimization_restarts >= 1:
            self.gpy_model.optimize_restarts(
                self.n_optimization_restarts, verbose=self.verbose_optimization, robust=True
            )

    @property
    def X(self) -> np.ndarray:
        return self.gpy_model.X

    @property
    def Y(self) -> np.ndarray:
        return self.gpy_model.Y

    def predict_covariance(self, X: np.ndarray, with_noise: bool = True) -> np.ndarray:
        """
        Calculates posterior covariance between points in X

        :param X: Array of size n_points x n_dimensions containing input locations to compute posterior covariance at
        :param with_noise: Whether to include likelihood noise in the covariance matrix
        :return: Posterior covariance matrix of size n_points x n_points
        """
        output_index = X[:, -1]
        y_metadata = {"output_index": output_index.astype(int)}
        variance = self.gpy_model.predict(X, Y_metadata=y_metadata, full_cov=True)[1]
        variance = np.maximum(variance, 1e-10)
        return variance

    def get_covariance_between_points(self, X1: np.ndarray, X2: np.ndarray) -> np.ndarray:
        """
        Calculate posterior covariance between two points
        :param X1: An array of shape 1 x n_dimensions that contains a data single point. It is the first argument of the
                   posterior covariance function
        :param X2: An array of shape n_points x n_dimensions that may contain multiple data points. This is the second
                   argument to the posterior covariance function.
        :return: An array of shape n_points x 1 of posterior covariances between X1 and X2
        """
        return self.gpy_model.posterior_covariance_between_points(X1, X2, include_likelihood=False)

    def generate_hyperparameters_samples(
        self, n_samples=10, n_burnin=5, subsample_interval=1, step_size=1e-1, leapfrog_steps=1
    ) -> np.ndarray:
        """
        Generates the samples from the hyper-parameters, and returns them (a numpy array whose rows are
        samples of the hyper-parameters).
        :param n_samples: Number of generated samples.
        :param n_burnin: Number of initial samples not used.
        :param subsample_interval: Interval of subsampling from HMC samples.
        :param step_size: Size of the gradient steps in the HMC sampler.
        :param leapfrog_steps: Number of gradient steps before each Metropolis Hasting step.
        """
        self.gpy_model.optimize(max_iters=self.n_optimization_restarts)
        self.gpy_model.param_array[:] = self.gpy_model.param_array * (
            1.0 + np.random.randn(self.gpy_model.param_array.size) * 0.01
        )
        hmc = GPy.inference.mcmc.HMC(self.gpy_model, stepsize=step_size)
        samples = hmc.sample(num_samples=n_burnin + n_samples * subsample_interval, hmc_iters=leapfrog_steps)
        return samples[n_burnin::subsample_interval]

    def fix_model_hyperparameters(self, sample_hyperparameters: np.ndarray) -> None:
        """
        Fix model hyperparameters

        """
        if self.gpy_model._fixes_ is None:
            self.gpy_model[:] = sample_hyperparameters
        else:
            self.gpy_model[self.gpy_model._fixes_] = sample_hyperparameters
        self.gpy_model._trigger_params_changed()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple

import numpy as np

from ...quadrature.interfaces.base_gp import IBaseGaussianProcess
from .warped_bq_model import WarpedBayesianQuadratureModel
from .warpings import IdentityWarping


class VanillaBayesianQuadrature(WarpedBayesianQuadratureModel):
    """Standard ('vanilla') Bayesian quadrature.

    The warping for vanilla Bayesian quadrature is the identity transform :math:`w(y) = y`.
    Hence, the model for the integrand :math:`f` is a standard Gaussian process as.

    :param base_gp: The underlying Gaussian process model.
    :param X: The initial locations of integrand evaluations, shape (n_points, input_dim).
    :param Y: The values of the integrand at X, shape (n_points, 1).

    """

    def __init__(self, base_gp: IBaseGaussianProcess, X: np.ndarray, Y: np.ndarray):
        super(VanillaBayesianQuadrature, self).__init__(base_gp=base_gp, warping=IdentityWarping(), X=X, Y=Y)

    def predict_base(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        m, cov = self.base_gp.predict(X_pred)
        return m, cov, m, cov

    def predict_base_with_full_covariance(
        self, X_pred: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        m, cov = self.base_gp.predict_with_full_covariance(X_pred)
        return m, cov, m, cov

    def integrate(self) -> Tuple[float, float]:
        kernel_mean_X = self.base_gp.kern.qK(self.X)
        integral_mean = np.dot(kernel_mean_X, self.base_gp.graminv_residual())[0, 0]
        integral_var = self.base_gp.kern.qKq() - (kernel_mean_X @ self.base_gp.solve_linear(kernel_mean_X.T))[0, 0]
        return integral_mean, integral_var

    def get_prediction_gradients(self, X: np.ndarray) -> Tuple:
        return self.base_gp.get_prediction_gradients(X)



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np

from ....core.loop.candidate_point_calculators import CandidatePointCalculator
from ....core.loop.loop_state import LoopState
from ....core.optimization.context_manager import ContextManager
from ....core.parameter_space import ParameterSpace
from ...methods.warped_bq_model import WarpedBayesianQuadratureModel


class BayesianMonteCarloPointCalculator(CandidatePointCalculator):
    """This point calculator produces Monte Carlo points from the integration measure.

    It can be used for Bayesian Monte Carlo (BMC) `[1]`_ as described in Section 2.1 of the paper.

    .. _[1]:

    .. rubric:: References

    [1] C.E. Rasmussen and Z. Ghahramani, Bayesian Monte Carlo,
    Advances in Neural Information Processing Systems 15 (NeurIPS) 2003

    .. seealso::
        :class:`emukit.quadrature.loop.BayesianMonteCarlo`

    :param model: A warped Bayesian quadrature model.
    :param parameter_space: The parameter space.

    :raises ValueError: If integration measure provided through the model does have sampling capabilities.

    """

    def __init__(self, model: WarpedBayesianQuadratureModel, parameter_space: ParameterSpace):
        self.parameter_space = parameter_space
        self.model = model

        # if measure is probability measure, check if it has sampling capabilities
        if self.model.base_gp.kern.measure is not None:
            if not self.model.base_gp.kern.measure.can_sample:
                raise ValueError(
                    "The given probability measure does not support sampling, but Bayesian Monte Carlo "
                    "requires sampling capability."
                )

    def compute_next_points(self, loop_state: LoopState, context: dict = None) -> np.ndarray:
        """Computes the next point.

        :param loop_state: Object that contains current state of the loop.
        :param context: Contains variables to fix and the values to fix them to. The dictionary key is the parameter
                        name and the value is the value to fix the parameter to.
        :return: The next point to evaluate the function at, shape (1, input_dim).
        """

        measure = self.model.base_gp.kern.measure

        # Lebesgue measure
        if measure is None:
            if context is None:
                return self.parameter_space.sample_uniform(1)
            else:
                context_manager = ContextManager(self.parameter_space, context)
                samples = context_manager.contextfree_space.sample_uniform(1)
                return context_manager.expand_vector(samples)

        # probability measure
        else:
            if context is None:
                return measure.sample(1)
            else:
                context_manager = ContextManager(self.parameter_space, context)
                samples = measure.sample(1, context_manager)
                return samples



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Optional, Tuple

import numpy as np

from .. import ParameterSpace
from ..acquisition import Acquisition
from .context_manager import Context, ContextManager


class AcquisitionOptimizerBase(abc.ABC):
    """
    Base class for acquisition optimizers
    """

    def __init__(self, space: ParameterSpace):
        """
        :param space: Parameter space containing entire input domain including any context variables
        """
        self.space = space

    @abc.abstractmethod
    def _optimize(self, acquisition: Acquisition, context_manager: ContextManager) -> Tuple[np.ndarray, np.ndarray]:
        """
        Implementation of optimization. See class docstring for details.

        :param acquisition: The acquisition function to be optimized
        :param context_manager: Optimization context manager.
        :return: Tuple of (location of maximum, acquisition value at maximizer)
        """
        pass

    def optimize(self, acquisition: Acquisition, context: Optional[Context] = None) -> Tuple[np.ndarray, np.ndarray]:
        """
        Optimizes the acquisition function.

        :param acquisition: The acquisition function to be optimized
        :param context: Optimization context.
                        Determines whether any variable values should be fixed during the optimization
        :return: Tuple of (location of maximum, acquisition value at maximizer)
        """
        if context is None:
            context = dict()
        context_manager = ContextManager(self.space, context)
        max_x, max_value = self._optimize(acquisition, context_manager)

        # Optimization might not match any encoding exactly
        # Rounding operation here finds the closest encoding
        rounded_max_x = self.space.round(max_x)

        if not np.array_equal(max_x, rounded_max_x):
            # re-evaluate if x changed while rounding to make sure value is correct
            return rounded_max_x, acquisition.evaluate(rounded_max_x)
        else:
            return max_x, max_value



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Callable, Union

import numpy as np
import scipy

from ...core import InformationSourceParameter
from ...core.acquisition import Acquisition
from ...core.interfaces import IModel
from ...core.parameter_space import ParameterSpace
from ...samplers import AffineInvariantEnsembleSampler, McmcSampler
from .. import epmgp
from ..acquisitions import ExpectedImprovement
from ..interfaces import IEntropySearchModel


class EntropySearch(Acquisition):
    def __init__(
        self,
        model: Union[IModel, IEntropySearchModel],
        space: ParameterSpace,
        sampler: McmcSampler = None,
        num_samples: int = 100,
        num_representer_points: int = 50,
        proposal_function: Callable = None,
        burn_in_steps: int = 50,
    ) -> None:

        """
        Entropy Search acquisition function approximates the distribution of the global
        minimum and tries to decrease its entropy. See this paper for more details:

        P. Hennig and C. J. Schuler
        Entropy search for information-efficient global optimization
        Journal of Machine Learning Research, 13, 2012

        :param model: GP model to compute the distribution of the minimum dubbed pmin.
        :param space: Domain space which we need for the sampling of the representer points
        :param sampler: mcmc sampler for representer points
        :param num_samples: integer determining how many samples to draw for each candidate input
        :param num_representer_points: integer determining how many representer points to sample
        :param proposal_function: Function that defines an unnormalized log proposal measure from which to sample the
        representer points. The default is expected improvement.
        :param burn_in_steps: integer that defines the number of burn-in steps when sampling the representer points
        """
        super().__init__()

        if not isinstance(model, IEntropySearchModel):
            raise RuntimeError("Model is not supported for Entropy Search")

        self.model = model
        self.num_representer_points = num_representer_points
        self.burn_in_steps = burn_in_steps

        if sampler is None:
            self.sampler = AffineInvariantEnsembleSampler(space)
        else:
            self.sampler = sampler

        # (unnormalized) density from which to sample the representer points to approximate pmin
        if proposal_function is None:

            ei = ExpectedImprovement(model)

            def prop_func(x):
                if len(x.shape) == 1:
                    x_ = x[None, :]
                else:
                    x_ = x

                if space.check_points_in_domain(x_):
                    return np.log(np.clip(ei.evaluate(x_)[0], 0.0, np.PINF))
                else:
                    return np.array([np.NINF])

            self.proposal_function = prop_func
        else:
            self.proposal_function = proposal_function

        # This is used later to calculate derivative of the stochastic part for the loss function
        # Derived following Ito's Lemma, see for example https://en.wikipedia.org/wiki/It%C3%B4%27s_lemma
        self.W = scipy.stats.norm.ppf(np.linspace(1.0 / (num_samples + 1), 1 - 1.0 / (num_samples + 1), num_samples))[
            np.newaxis, :
        ]

        # Initialize parameters to lazily compute them once needed
        self.representer_points = None
        self.representer_points_log = None
        self.logP = None
        self.p_min_entropy = None

    def _sample_representer_points(self) -> tuple:
        """Samples a new set of representer points from the proposal measurement"""

        repr_points, repr_points_log = self.sampler.get_samples(
            self.num_representer_points, self.proposal_function, self.burn_in_steps
        )

        if np.any(np.isnan(repr_points_log)) or np.any(np.isposinf(repr_points_log)):
            raise RuntimeError(
                "Sampler generated representer points with invalid log values: {}".format(repr_points_log)
            )

        # Removing representer points that have 0 probability of being the minimum
        idx_to_remove = np.where(np.isneginf(repr_points_log))[0]
        if len(idx_to_remove) > 0:
            idx = list(set(range(self.num_representer_points)) - set(idx_to_remove))
            repr_points = repr_points[idx, :]
            repr_points_log = repr_points_log[idx]

        return repr_points, repr_points_log

    def update_parameters(self) -> None:
        """
        Updates p_min parameter
        """
        self.update_pmin()

    def update_pmin(self) -> np.ndarray:
        """
        Approximates the distribution of the global optimum  p(x=x_star|D) by doing the following steps:
            - discretizing the input space by representer points sampled from a proposal measure (default EI)
            - predicting mean and the covariance matrix of these representer points
            - uses EPMGP algorithm to compute the probability of each representer point being the minimum
        """

        self.representer_points, self.representer_points_log = self._sample_representer_points()

        mu, _ = self.model.predict(self.representer_points)
        mu = np.ndarray.flatten(mu)
        var = self.model.predict_covariance(self.representer_points)

        self.logP, self.dlogPdMu, self.dlogPdSigma, self.dlogPdMudMu = epmgp.joint_min(mu, var, with_derivatives=True)
        self.logP = self.logP[:, np.newaxis]

        # Calculate the entropy of the distribution over the minimum given the current model
        self.p_min_entropy = np.sum(
            np.multiply(np.exp(self.logP), np.add(self.logP, self.representer_points_log)), axis=0
        )

        return self.logP

    def _required_parameters_initialized(self):
        """
        Checks if all required parameters are initialized.
        """
        return not (self.representer_points is None or self.representer_points_log is None or self.logP is None)

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the information gain, i.e the change in entropy of p_min if we would evaluate x.

        :param x: points where the acquisition is evaluated.
        """
        if not self._required_parameters_initialized():
            self.update_pmin()

        # Check if we want to compute the acquisition function for multiple inputs
        if x.shape[0] > 1:
            results = np.zeros([x.shape[0], 1])
            for j in range(x.shape[0]):
                results[j] = self.evaluate(x[j, None, :])
            return results

        # Number of representer points locations
        N = self.logP.size

        # Evaluate innovations, i.e how much does mean and variance at the
        # representer points change if we would evaluate x
        dMdx, dVdx = self._innovations(x)

        dVdx = dVdx[np.triu(np.ones((N, N))).T.astype(bool), np.newaxis]

        dMdx_squared = dMdx.dot(dMdx.T)
        trace_term = np.sum(
            np.sum(
                np.multiply(
                    self.dlogPdMudMu, np.reshape(dMdx_squared, (1, dMdx_squared.shape[0], dMdx_squared.shape[1]))
                ),
                2,
            ),
            1,
        )[:, np.newaxis]

        # Deterministic part of change:
        deterministic_change = self.dlogPdSigma.dot(dVdx) + 0.5 * trace_term
        # Stochastic part of change:
        stochastic_change = (self.dlogPdMu.dot(dMdx)).dot(self.W)

        # Update our pmin distribution
        predicted_logP = np.add(self.logP + deterministic_change, stochastic_change)
        max_predicted_logP = np.amax(predicted_logP, axis=0)

        # normalize predictions
        max_diff = max_predicted_logP + np.log(np.sum(np.exp(predicted_logP - max_predicted_logP), axis=0))
        lselP = max_predicted_logP if np.any(np.isinf(max_diff)) else max_diff
        predicted_logP = np.subtract(predicted_logP, lselP)

        # We maximize the information gain
        H_p = np.sum(np.multiply(np.exp(predicted_logP), np.add(predicted_logP, self.representer_points_log)), axis=0)

        new_entropy = np.mean(H_p)
        entropy_change = new_entropy - self.p_min_entropy
        return entropy_change.reshape(-1, 1)

    def _innovations(self, x: np.ndarray) -> tuple:
        """
        Computes the expected change in mean and variance at the representer
        points (cf. Section 2.4 in the paper).

        :param x: candidate for which to compute the expected change in the GP
        """

        # Get the standard deviation at x without noise
        stdev_x = np.sqrt(self.model.predict_covariance(x, with_noise=False))

        # Compute the variance between the test point x and the representer points
        sigma_x_rep = self.model.get_covariance_between_points(self.representer_points, x)
        dm_rep = sigma_x_rep / stdev_x

        # Compute the deterministic innovation for the variance
        dv_rep = -dm_rep.dot(dm_rep.T)
        return dm_rep, dv_rep

    @property
    def has_gradients(self) -> bool:
        """Returns that this acquisition has gradients"""
        return False


class MultiInformationSourceEntropySearch(EntropySearch):
    """
    Entropy search acquisition for multi-information source problems where the objective function is the output of one
    of the information sources. The other information sources provide auxiliary information about the objective function
    """

    def __init__(
        self,
        model: Union[IModel, IEntropySearchModel],
        space: ParameterSpace,
        target_information_source_index: int = None,
        num_samples: int = 100,
        num_representer_points: int = 50,
        burn_in_steps: int = 50,
    ):
        """
        :param model: Gaussian process model of the objective function that implements IEntropySearchModel
        :param space: Parameter space of the input domain
        :param target_information_source_index: The index of the information source we want to minimise
        :param num_samples: Integer determining how many samples to draw for each candidate input
        :param num_representer_points: Integer determining how many representer points to sample
        :param burn_in_steps: Integer that defines the number of burn-in steps when sampling the representer points
        """

        # Find information source parameter in parameter space
        info_source_parameter, source_idx = _find_source_parameter(space)
        self.source_idx = source_idx

        # Assume we are in a multi-fidelity setting and the highest index is the highest fidelity
        if target_information_source_index is None:
            target_information_source_index = max(info_source_parameter.domain)
        self.target_information_source_index = target_information_source_index

        # Sampler of representer points should sample x location at the target information source only so make a
        # parameter space without the information source parameter
        parameters_without_info_source = space.parameters.copy()
        parameters_without_info_source.remove(info_source_parameter)
        space_without_info_source = ParameterSpace(parameters_without_info_source)

        # Create sampler of representer points
        sampler = AffineInvariantEnsembleSampler(space_without_info_source)

        proposal_func = self._get_proposal_function(model, space)

        super().__init__(model, space, sampler, num_samples, num_representer_points, proposal_func, burn_in_steps)

    def _sample_representer_points(self):
        repr_points, repr_points_log = super()._sample_representer_points()

        # Add fidelity index to representer points
        idx = np.ones((repr_points.shape[0])) * self.target_information_source_index
        repr_points = np.insert(repr_points, self.source_idx, idx, axis=1)
        return repr_points, repr_points_log

    def _get_proposal_function(self, model, space):

        # Define proposal function for multi-fidelity
        ei = ExpectedImprovement(model)

        def proposal_func(x):
            x_ = x[None, :]

            # Add information source parameter into array
            idx = np.ones((x_.shape[0], 1)) * self.target_information_source_index
            x_ = np.insert(x_, self.source_idx, idx, axis=1)

            if space.check_points_in_domain(x_):
                val = np.log(np.clip(ei.evaluate(x_)[0], 0.0, np.PINF))
                if np.any(np.isnan(val)):
                    return np.array([np.NINF])
                else:
                    return val
            else:
                return np.array([np.NINF])

        return proposal_func


def _find_source_parameter(space):
    # Find information source parameter in parameter space
    info_source_parameter = None
    source_idx = None
    for i, param in enumerate(space.parameters):
        if isinstance(param, InformationSourceParameter):
            info_source_parameter = param
            source_idx = i

    if info_source_parameter is None:
        raise ValueError("No information source parameter found in the parameter space")

    return info_source_parameter, source_idx



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np

from ...bayesian_optimization.acquisitions import ExpectedImprovement
from ...core.acquisition import Acquisition, acquisition_per_expected_cost
from ...core.interfaces import IModel
from ...core.loop import FixedIntervalUpdater, OuterLoop, SequentialPointCalculator
from ...core.loop.loop_state import create_loop_state
from ...core.optimization import AcquisitionOptimizerBase, GradientAcquisitionOptimizer
from ...core.parameter_space import ParameterSpace


class CostSensitiveBayesianOptimizationLoop(OuterLoop):
    def __init__(
        self,
        space: ParameterSpace,
        model_objective: IModel,
        model_cost: IModel,
        acquisition: Acquisition = None,
        update_interval: int = 1,
        acquisition_optimizer: AcquisitionOptimizerBase = None,
    ):

        """
        Emukit class that implements a loop for building modular cost sensitive Bayesian optimization.

        :param space: Input space where the optimization is carried out.
        :param model_objective: The model that approximates the underlying objective function
        :param model_cost: The model that approximates the cost of evaluating the objective function
        :param acquisition: The acquisition function that will be used to collect new points (default, EI).
        :param update_interval:  Number of iterations between optimization of model hyper-parameters. Defaults to 1.
        :param acquisition_optimizer: Optimizer selecting next evaluation points
                                      by maximizing acquisition.
                                      Gradient based optimizer is used if None.
                                      Defaults to None.
        """

        if not np.all(np.isclose(model_objective.X, model_cost.X)):
            raise ValueError(
                "Emukit currently only supports identical " "training inputs for the cost and objective model"
            )

        if acquisition is None:
            expected_improvement = ExpectedImprovement(model_objective)
            acquisition = acquisition_per_expected_cost(expected_improvement, model_cost)

        model_updater_objective = FixedIntervalUpdater(model_objective, update_interval)
        model_updater_cost = FixedIntervalUpdater(model_cost, update_interval, lambda state: state.cost)

        if acquisition_optimizer is None:
            acquisition_optimizer = GradientAcquisitionOptimizer(space)
        candidate_point_calculator = SequentialPointCalculator(acquisition, acquisition_optimizer)

        loop_state = create_loop_state(model_objective.X, model_objective.Y, cost=model_cost.Y)

        super(CostSensitiveBayesianOptimizationLoop, self).__init__(
            candidate_point_calculator, [model_updater_objective, model_updater_cost], loop_state
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple

import numpy as np

from ...core.interfaces import IDifferentiable, IModel
from ..kernels.quadrature_kernels import QuadratureKernel


class IBaseGaussianProcess(IModel, IDifferentiable):
    """Interface for a Gaussian process as used by quadrature models.

    Implementations of this class can be used by Gaussian process based quadrature models
    such as :class:`WarpedBayesianQuadratureModel`.

    .. note::
        When this class is initialized with data, use the raw evaluations Y of the integrand
        as this is a general Gaussian process class that can be used with various quadrature methods.
        Possible transformations of Y such as the ones applied in :class:`WarpedBayesianQuadratureModel`
        will be performed automatically there.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IStandardKernel`
       * :class:`emukit.quadrature.methods.WarpedBayesianQuadratureModel`

    :param kern: An instance of a quadrature kernel.
    """

    def __init__(self, kern: QuadratureKernel) -> None:
        self.kern = kern

    @property
    def observation_noise_variance(self) -> float:
        """The variance of the Gaussian observation noise."""
        raise NotImplementedError

    def predict_with_full_covariance(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Predictive mean and full co-variance at the locations X_pred.

        :param X_pred: Points at which to predict, shape (n_points, input_dim)
        :return: Predictive mean, predictive full co-variance, shapes (num_points, 1) and (num_points, num_points).
        """
        raise NotImplementedError

    def solve_linear(self, z: np.ndarray) -> np.ndarray:
        """Solve the linear system :math:`Gx=z` for :math:`x`.

        :math:`G` is the Gram matrix :math:`G := K(X, X) + \sigma^2 I`,
        of shape (num_dat, num_dat) and :math:`z` is a matrix of shape (num_dat, num_obs).

        :param z: A matrix of shape (num_dat, num_obs).
        :return: The solution :math:`x` of the linear, shape (num_dat, num_obs).
        """
        raise NotImplementedError

    def graminv_residual(self) -> np.ndarray:
        r"""The solution :math:`z` of the linear system

        .. math::

            (K_{XX} + \zeta^2 I) z = (Y - m(X))

        where :math:`X` and :math:`Y` are the available nodes and function evaluation, :math:`m(X)`
        is the predictive mean at :math:`X`, and :math:`\zeta^2` the observation noise variance.

        :return: The solution :math:`z` of the linear system, shape (num_dat, 1).
        """
        raise NotImplementedError

    def get_prediction_gradients(self, X: np.ndarray) -> Tuple:
        """Evaluates the predictive gradients of mean and variance at X.

        :param X: Points at which the gradients are evaluated, shape (n_points, input_dim).
        :returns: The gradients of mean and variance, each of shape (n_points, input_dim).
        """
        # gradient of mean
        d_mean_dx = (self.kern.dK_dx1(X, self.X) @ self.graminv_residual())[:, :, 0].T

        # gradient of variance
        dKdiag_dx = self.kern.dKdiag_dx(X)
        dKxX_dx1 = self.kern.dK_dx1(X, self.X)
        graminv_KXx = self.solve_linear(self.kern.K(self.X, X))
        d_var_dx = dKdiag_dx - 2.0 * (dKxX_dx1 * np.transpose(graminv_KXx)).sum(axis=2, keepdims=False)

        return d_mean_dx, d_var_dx.T



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List, Tuple, Union

import numpy as np

from .parameter import Parameter


class ContinuousParameter(Parameter):
    """
    A univariate continuous parameter with a domain defined in a range between two values
    """

    def __init__(self, name: str, min_value: float, max_value: float):
        """
        :param name: Name of parameter
        :param min_value: Minimum value the parameter is allowed to take
        :param max_value: Maximum value the parameter is allowed to take
        """
        super().__init__(name)
        self.min = min_value
        self.max = max_value

    def __str__(self):
        return f"<ContinuousParameter: {self.name} {self.bounds}>"

    def __repr__(self):
        return f"ContinuousParameter({self.name}, {self.min}, {self.max})"

    def check_in_domain(self, x: Union[np.ndarray, float]) -> bool:
        """
        Checks if all the points in x lie between the min and max allowed values

        :param x:    1d numpy array of points to check
                  or 2d numpy array with shape (n_points, 1) of points to check
                  or float of single point to check
        :return: A boolean value which indicates whether all points lie in the domain
        """
        if isinstance(x, np.ndarray):
            if x.ndim == 2 and x.shape[1] == 1:
                x = x.ravel()
            elif x.ndim > 1:
                raise ValueError("Expected x shape (n,) or (n, 1), actual is {}".format(x.shape))
        return np.all([self.min <= x, x <= self.max])

    @property
    def bounds(self) -> List[Tuple]:
        """
        Returns a list containing one tuple of minimum and maximum values parameter can take
        """
        return [(self.min, self.max)]

    def sample_uniform(self, point_count: int) -> np.ndarray:
        """
        Generates multiple uniformly distributed random parameter points.

        :param point_count: number of data points to generate.
        :returns: Generated points with shape (point_count, num_features)
        """
        return np.random.uniform(low=self.min, high=self.max, size=(point_count, 1))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Tuple

import numpy as np

from .. import ParameterSpace
from ..acquisition import Acquisition
from .acquisition_optimizer import AcquisitionOptimizerBase
from .anchor_points_generator import ObjectiveAnchorPointsGenerator
from .context_manager import ContextManager
from .optimizer import OptLbfgs, OptTrustRegionConstrained, apply_optimizer

_log = logging.getLogger(__name__)


class GradientAcquisitionOptimizer(AcquisitionOptimizerBase):
    """Optimizes the acquisition function using a quasi-Newton method (L-BFGS).
    Can be used for continuous acquisition functions.
    """

    def __init__(self, space: ParameterSpace, num_samples=1000, num_anchor=1) -> None:
        """
        :param space: The parameter space spanning the search problem.
        """
        super().__init__(space)
        self.num_samples = num_samples
        self.num_anchor = num_anchor

    def _optimize(self, acquisition: Acquisition, context_manager: ContextManager) -> Tuple[np.ndarray, np.ndarray]:
        """
        Implementation of abstract method.
        Taking into account gradients if acquisition supports them.

        See AcquisitionOptimizerBase._optimizer for parameter descriptions.
        See class docstring for implementation details.
        """

        # Take negative of acquisition function because they are to be maximised and the optimizers minimise
        f = lambda x: -acquisition.evaluate(x)

        # Context validation
        if len(context_manager.contextfree_space.parameters) == 0:
            _log.warning("All parameters are fixed through context")
            x = np.array(context_manager.context_values)[None, :]
            return x, f(x)

        if acquisition.has_gradients:

            def f_df(x):
                f_value, df_value = acquisition.evaluate_with_gradients(x)
                return -f_value, -df_value

        else:
            f_df = None

        optimizer = self._get_optimizer(context_manager)
        anchor_points_generator = ObjectiveAnchorPointsGenerator(self.space, acquisition, num_samples=self.num_samples)

        # Select the anchor points (with context)
        anchor_points = anchor_points_generator.get(num_anchor=self.num_anchor, context_manager=context_manager)

        _log.info("Starting gradient-based optimization of acquisition function {}".format(type(acquisition)))
        optimized_points = []
        for a in anchor_points:
            optimized_point = apply_optimizer(
                optimizer, a, space=self.space, f=f, df=None, f_df=f_df, context_manager=context_manager
            )
            optimized_points.append(optimized_point)

        x_min, fx_min = min(optimized_points, key=lambda t: t[1])
        return x_min, -fx_min

    def _get_optimizer(self, context_manager):
        if len(self.space.constraints) == 0:
            return OptLbfgs(context_manager.contextfree_space.get_bounds())
        else:
            return OptTrustRegionConstrained(context_manager.contextfree_space.get_bounds(), self.space.constraints)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple

import numpy as np

from ...core.acquisition import Acquisition
from ..methods import VanillaBayesianQuadrature


class SquaredCorrelation(Acquisition):
    r"""The squared correlation between the integral value and integrand evaluations
    under a Gaussian process model.

    .. math::
        a(x) = \frac{(\int k_N(s, x)p(s)\mathrm{d}s)^2}{\mathfrak{v}_N v_N(x)}\in [0, 1]

    where :math:`\mathfrak{v}_N` is the current integral variance given :math:`N` observations,
    :math:`v_N(x)` is the predictive integral variance if point :math:`x` was added newly,
    and :math:`k_N(s, x)` is the posterior kernel function.

    .. note::
        The squared correlation is identical to the integral-variance-reduction acquisition up to a
        global normalizing constant under a standard Gaussian process model.

    .. seealso::
        :class:`emukit.quadrature.acquisitions.IntegralVarianceReduction`

    :param model: A vanilla Bayesian quadrature model.

    """

    def __init__(self, model: VanillaBayesianQuadrature):
        self.model = model

    @property
    def has_gradients(self) -> bool:
        """Whether acquisition value has analytical gradient calculation available."""
        return True

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """Evaluates the acquisition function at x.

        :param x: The locations where to evaluate, shape (n_points, input_dim) .
        :return: The acquisition values at x, shape (n_points, 1).
        """
        return self._evaluate(x)[0]

    def _evaluate(self, x: np.ndarray) -> Tuple[np.ndarray, float, np.ndarray, np.ndarray]:
        """Evaluates the acquisition function at x.

        :param x: The locations where to evaluate, shape (n_points, input_dim) .
        :return: The acquisition values at x (n_points, 1),
                 the current integral variance (float),
                 the predictive variance + noise at x (n_points, 1),
                 the predictive covariance between integral and x (n_points, 1).
        """
        integral_current_var, y_predictive_var, predictive_cov = self._value_terms(x)
        squared_correlation = predictive_cov**2 / (integral_current_var * y_predictive_var)
        return squared_correlation, integral_current_var, y_predictive_var, predictive_cov

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Evaluate the acquisition function and its gradient.

        :param x: The locations where to evaluate, shape (n_points, input_dim).
        :return: The acquisition values and corresponding gradients at x,
                 shapes (n_points, 1) and (n_points, input_dim)
        """
        # value
        squared_correlation, integral_current_var, y_predictive_var, predictive_cov = self._evaluate(x)

        # gradient
        d_y_predictive_var_dx, d_predictive_cov_dx = self._gradient_terms(x)
        first_term = 2.0 * predictive_cov * d_predictive_cov_dx
        second_term = (predictive_cov**2 / y_predictive_var) * d_y_predictive_var_dx
        normalization = integral_current_var * y_predictive_var
        squared_correlation_gradient = (first_term - second_term) / normalization

        return squared_correlation, squared_correlation_gradient

    def _value_terms(self, x: np.ndarray) -> Tuple[float, np.ndarray, np.ndarray]:
        """Computes the terms needed for the squared correlation.

        :param x: The locations where to evaluate, shape (n_points, input_dim).
        :return: The current integral variance (float),
                 predictive variance + noise at x (n_points, 1),
                 predictive covariance between integral and x (n_points, 1).
        """
        integral_current_var = self.model.integrate()[1]
        y_predictive_var = self.model.predict(x)[1] + self.model.base_gp.observation_noise_variance

        qKx = self.model.base_gp.kern.qK(x)
        qKX = self.model.base_gp.kern.qK(self.model.base_gp.X)

        graminv_KXx = self.model.base_gp.solve_linear(self.model.base_gp.kern.K(self.model.base_gp.X, x))
        predictive_cov = np.transpose(qKx - np.dot(qKX, graminv_KXx))
        return integral_current_var, y_predictive_var, predictive_cov

    def _gradient_terms(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Computes the terms needed for the gradient of the squared correlation.

        :param x: The locations where to evaluate, shape (n_points, input_dim).
        :return: The gradient of y_predictive_var and predictive_cov wrt. x at x, shapes (n_points, input_dim).
        """
        # gradient of predictive variance of y
        d_y_predictive_var_dx = self.model.get_prediction_gradients(x)[1].T

        # gradient of predictive covariance between integral and (x, y)-pair
        dqKx_dx = np.transpose(self.model.base_gp.kern.dqK_dx(x))

        qKX = self.model.base_gp.kern.qK(self.model.base_gp.X)
        qKX_graminv = np.transpose(self.model.base_gp.solve_linear(qKX.T))  # (1, N)

        dKXx_dx2 = self.model.base_gp.kern.dK_dx2(self.model.X, x)
        d_predictive_cov_dx = dqKx_dx - np.dot(qKX_graminv, np.transpose(dKXx_dx2))[0, :, :]

        return np.transpose(d_y_predictive_var_dx), d_predictive_cov_dx



"""Interfaces for the quadrature package."""

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .base_gp import IBaseGaussianProcess  # noqa: F401
from .standard_kernels import (  # noqa: F401
    IRBF,
    IBrownian,
    IProductBrownian,
    IProductMatern12,
    IProductMatern32,
    IProductMatern52,
    IStandardKernel,
)

__all__ = [
    "IBaseGaussianProcess",
    "IStandardKernel",
    "IBrownian",
    "IRBF",
    "IProductBrownian",
    "IProductMatern12",
    "IProductMatern32",
    "IProductMatern52",
]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

import numpy as np

from ...core.optimization.context_manager import ContextManager
from ..typing import BoundsType
from .domain import BoxDomain


class IntegrationMeasure:
    r"""An abstract class for an integration measure defined by a density.

    :param domain: The domain. ``None`` implies :math:`\mathbb{R}^d`.
    :param name: Name of the integration measure

    """

    def __init__(self, domain: Optional[BoxDomain], name: str):
        self.domain = domain
        self.name = name

    @property
    def input_dim(self):
        """The input dimensionality."""
        raise NotImplementedError

    @property
    def can_sample(self) -> bool:
        """Indicates whether the measure has sampling available.

        :return: ``True`` if sampling is available. ``False`` otherwise.
        """
        raise NotImplementedError

    def compute_density(self, x: np.ndarray) -> np.ndarray:
        """Evaluates the density at x.

        :param x: Points at which density is evaluated, shape (n_points, input_dim).
        :return: The density at x, shape (n_points, ).
        """
        raise NotImplementedError

    def compute_density_gradient(self, x: np.ndarray) -> np.ndarray:
        """Evaluates the gradient of the density at x.

        :param x: Points at which the gradient is evaluated, shape (n_points, input_dim).
        :return: The gradient of the density at x, shape (n_points, input_dim).
        """
        raise NotImplementedError

    def reasonable_box(self) -> BoundsType:
        """A reasonable box containing the measure.

        Outside this box, the measure should be zero or virtually zero.

        :return: The reasonable box.
        """
        raise NotImplementedError

    def sample(self, num_samples: int, context_manager: ContextManager = None) -> np.ndarray:
        """Samples from the measure.

        :param num_samples: The number of samples to be taken.
        :param context_manager: The context manager that contains variables to fix and the values to fix them to.
                                If a context is given, this method samples from the conditional distribution.
        :return: The samples, shape (num_samples, input_dim).
        """
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from ...core.acquisition import Acquisition
from ...core.loop import FixedIntervalUpdater, ModelUpdater, OuterLoop, SequentialPointCalculator
from ...core.loop.loop_state import create_loop_state
from ...core.optimization import AcquisitionOptimizerBase, GradientAcquisitionOptimizer
from ...core.parameter_space import ParameterSpace
from ..acquisitions import IntegralVarianceReduction
from ..methods import VanillaBayesianQuadrature


class VanillaBayesianQuadratureLoop(OuterLoop):
    """The loop for standard ('vanilla') Bayesian Quadrature.

    .. seealso::
        :class:`emukit.quadrature.methods.VanillaBayesianQuadrature`

    :param model: The vanilla Bayesian quadrature model.
    :param acquisition: The acquisition function that is used to collect new points,
                        defaults to integral-variance-reduction.
    :param model_updater: Defines how and when the BQ model is updated if new data arrives.
                          Defaults to updating hyper-parameters after every iteration.
    :param acquisition_optimizer: Optimizer selecting next evaluation points by maximizing acquisition.
                                  Gradient based optimizer is used if None. Defaults to None.
    """

    def __init__(
        self,
        model: VanillaBayesianQuadrature,
        acquisition: Acquisition = None,
        model_updater: ModelUpdater = None,
        acquisition_optimizer: AcquisitionOptimizerBase = None,
    ):
        if acquisition is None:
            acquisition = IntegralVarianceReduction(model)

        if model_updater is None:
            model_updater = FixedIntervalUpdater(model, 1)

        space = ParameterSpace(model.reasonable_box_bounds.convert_to_list_of_continuous_parameters())
        if acquisition_optimizer is None:
            acquisition_optimizer = GradientAcquisitionOptimizer(space)
        candidate_point_calculator = SequentialPointCalculator(acquisition, acquisition_optimizer)
        loop_state = create_loop_state(model.X, model.Y)

        super().__init__(candidate_point_calculator, model_updater, loop_state)

        self.model = model



"""Loops for Bayesian quadrature."""

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .bayesian_monte_carlo_loop import BayesianMonteCarlo  # noqa: F401
from .vanilla_bq_loop import VanillaBayesianQuadratureLoop  # noqa: F401
from .wsabil_loop import WSABILLoop  # noqa: F401

__all__ = [
    "BayesianMonteCarlo",
    "VanillaBayesianQuadratureLoop",
    "WSABILLoop",
    "point_calculators",
]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List, Tuple

import numpy as np

from emukit.core.loop.user_function import UserFunctionWrapper


def circular_gaussian(
    mean: float = 0.0, variance: float = 1.0
) -> Tuple[UserFunctionWrapper, List[Tuple[float, float]]]:
    r"""2D toy integrand that is a Gaussian on a circle.

    .. math::
        f(x) = (2\pi \sigma^2)^{-\frac{1}{2}} r^2 e^{-\frac{(r - \mu)^2}{2 \sigma^2}}

    where :math:`\sigma^2` is the :attr:`variance` attribute,
    :math:`\mu` is the :attr:`mean` attribute
    and :math:`r = \|x\|` is the length of the input :math:`x`.

    :param mean: The mean of the circular Gaussian in units of radius (must be >= 0, defaults to 0).
    :param variance: The variance of the Gaussian (must be > 0, defaults to 1.).
    :return: The wrapped test function, and the integrals bounds (the latter defaults to [-3, 3]^2).
    """
    func = lambda x: _circular_gaussian(x, mean, variance)
    integral_bounds = 2 * [(-3.0, 3.0)]
    return UserFunctionWrapper(func), integral_bounds


def _circular_gaussian(x: np.ndarray, mean: float, variance: float) -> np.ndarray:
    """
    :param x: Locations for evaluation (num_points, 2).
    :param mean: The mean of Gaussian in units of radius (must be >= 0).
    :param variance: The variance of the Gaussian (must be > 0).
    :return: The function values at x, shape (num_points, 1).
    """
    norm_x = np.sqrt((x**2).sum(axis=1, keepdims=True))
    return norm_x**2 * np.exp(-((norm_x - mean) ** 2) / (2.0 * variance)) / np.sqrt(2.0 * np.pi * variance)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np


class Ishigami(object):
    """
    Ishigami function was first introduced in 'Ishigami and T. Homma, An importance quantification
    technique in uncertainty analysis for computer 556 models, in Uncertainty Modeling and Analysis,
    1990. Proceedings., First International Symposium 557 on, IEEE, 1990, pp. 398–403'. This class
    contains: the elements of the Sobol decomposition of the Ishigami function, three low fidelity
    approximations, and the values of the main effects of the Sobol components:

    .. math::

        f(x_1,x_2,x_3) = \sum f_i(x_i) + \sum f_{ij}(x_i,x_j) + f(x_1,x_2,x_3)

    """

    def __init__(self, a: float, b: float):
        """
        :param a: Constant in Ishigami equation. See method docstrings for details
        :param b: Constant in Ishigami equation. See method docstrings for details
        """
        self.a = a
        self.b = b

        self.variance_total = 0.5 + self.a**2 / 8 + np.pi**4 * self.b / 5 + np.pi**8 * self.b**2 / 18
        self.variance_x1 = 0.5 * (1 + self.b * np.pi**4 / 5) ** 2
        self.variance_x2 = self.a**2 / 8
        self.variance_x3 = 0
        self.variance_x12 = 0
        self.variance_x13 = np.pi**8 * self.b**2 * (1 / 18 - 1 / 50)
        self.variance_x23 = 0
        self.variance_x123 = 0

        # Main effects
        self.main_effects = {
            "x1": self.variance_x1 / self.variance_total,
            "x2": self.variance_x2 / self.variance_total,
            "x3": self.variance_x3 / self.variance_total,
        }

        # Total effects
        self.total_effects = {
            "x1": (self.variance_x1 + self.variance_x13) / self.variance_total,
            "x2": self.variance_x2 / self.variance_total,
            "x3": (self.variance_x3 + self.variance_x13) / self.variance_total,
        }

    def fidelity1(self, x: np.ndarray) -> np.ndarray:
        """
        Highest fidelity

        .. math::
            \sin(x_1) + a \sin(x_2)^2 + b x_3 ^ 4 \sin(x_1)

        :param x: (n_points x 3) numpy array containing inputs to function
        """
        return np.sin(x[:, 0]) + self.a * np.sin(x[:, 1]) ** 2 + self.b * x[:, 2] ** 4 * (np.sin(x[:, 0]))

    def fidelity2(self, x: np.ndarray) -> np.ndarray:
        """
        Low fidelity approximation

        .. math::
            \sin(x_1) + a \sin(x_2) + b x_3 ^ 4 \sin(x_1)

        :param x: (n_points x 3) numpy array containing inputs to function
        """
        return np.sin(x[:, 0]) + self.a * np.sin(x[:, 1]) + self.b * x[:, 2] ** 4 * (np.sin(x[:, 0]))

    def fidelity3(self, x: np.ndarray) -> np.ndarray:
        """
        Low fidelity approximation

        .. math::
            \sin(x_1) + 0.95 a \sin(x_2)^2 + b x_3 ^ 4 \sin(x_1)

        :param x: (n_points x 3) numpy array containing inputs to function
        """
        return np.sin(x[:, 0]) + 0.95 * self.a * np.sin(x[:, 1]) ** 2 + self.b * x[:, 2] ** 4 * (np.sin(x[:, 0]))

    def fidelity4(self, x: np.ndarray) -> np.ndarray:
        """
        Low fidelity approximation

        .. math::
            \sin(x_1) + 0.6 a \sin(x_2)^2 + 9 b x_3 ^ 4 \sin(x_1)

        :param x: (n_points x 3) numpy array containing inputs to function
        """
        return np.sin(x[:, 0]) + 0.6 * self.a * np.sin(x[:, 1]) ** 2 + 9 * self.b * x[:, 2] ** 4 * (np.sin(x[:, 0]))

    def f0(self) -> float:
        """
        Constant effect of high fidelity function
        """
        return self.a / 2

    def f1(self, x1) -> np.ndarray:
        """
        First order effect of x_1 on high fidelity function
        """
        return (1 + (self.b / 5) * np.pi**4) * np.sin(x1)

    def f2(self, x2) -> np.ndarray:
        """
        First order effect of x_2 on high fidelity function
        """
        return self.a * np.sin(x2) ** 2 - 0.5 * self.a

    def f3(self, x3) -> np.ndarray:
        """
        First order effect of x_3 on high fidelity function
        """
        return x3[:, 0] * 0

    def f12(self, x12) -> np.ndarray:
        """
        Second order effect of x_1 and x_2 on high fidelity function
        """
        return x12[:, 0] * 0

    def f13(self, x13) -> np.ndarray:
        """
        Second order effect of x_1 and x_3 on high fidelity function
        """
        return self.b * np.sin(x13[:, 0]) * (x13[:, 1] ** 4 - np.pi**4 / 5)

    def f23(self, x23) -> np.ndarray:
        """
        Second order effect of x_2 and x_3 on high fidelity function
        """
        return x23[:, 0] * 0

    def f123(self, x123) -> np.ndarray:
        """
        Third order effect of all inputs on high fidelity function
        """
        return x123[:, 0] * 0



"""Bayesian quadrature models."""

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .bounded_bq_model import BoundedBayesianQuadrature  # noqa: F401
from .vanilla_bq import VanillaBayesianQuadrature  # noqa: F401
from .warped_bq_model import WarpedBayesianQuadratureModel  # noqa: F401
from .wsabi import WSABIL  # noqa: F401

__all__ = [
    "WarpedBayesianQuadratureModel",
    "VanillaBayesianQuadrature",
    "BoundedBayesianQuadrature",
    "WSABIL",
    "warpings",
]



"""WSABI model as in Gunter et al. 2014"""


import numpy as np

from ..interfaces.base_gp import IBaseGaussianProcess
from . import BoundedBayesianQuadrature


class WSABIL(BoundedBayesianQuadrature):
    r"""Warped Sequential Active Bayesian Integration with linear approximation (WSABI-L).

    The linear approximation is described in `[1]`_ in section 3.1, Eq (9) and (10).

    The offset :math:`\alpha` (notation from paper) will either be set to a small value if
    ``adapt_alpha`` is ``False``. Else it will be adapted according to
    :math:`0.8 \operatorname{min}(Y)` as in Gunter et al. 2014, page 3, footnote,
    where :math:`Y` are the collected integrand evaluations so far.

    .. _[1]:

    .. rubric:: References

    [1] Gunter et al. 2014 *Sampling for Inference in Probabilistic Models with Fast Bayesian Quadrature*,
    Advances in Neural Information Processing Systems (NeurIPS), 27, pp. 2789–2797.

    .. note::
        WSABI-L must be used with the RBF kernel and the Gaussian integration measure
        (See description of :attr:`base_gp` parameter). The loop must use the uncertainty
        sampling acquisition strategy.

    .. seealso::
        * :class:`emukit.quadrature.methods.BoundedBayesianQuadrature`
        * :class:`emukit.quadrature.acquisitions.UncertaintySampling`
        * :class:`emukit.quadrature.loop.WSABILLoop`

    :param base_gp: A standard Gaussian process.
                    Must use :class:`emukit.quadrature.kernels.QuadratureRBFGaussianMeasure` as kernel.
    :param X: The initial locations of integrand evaluations, shape (n_points, input_dim).
    :param Y: The values of the integrand at X, shape (n_points, 1).
    :param adapt_alpha: If ``True``, the offset :math:`\alpha` will be adapted. If ``False`` :math:`\alpha` will be
                        fixed to a small value for numerical stability. Default is ``True``.

    """

    def __init__(self, base_gp: IBaseGaussianProcess, X: np.ndarray, Y: np.ndarray, adapt_alpha: bool = True):
        self.adapt_alpha = adapt_alpha
        self._small_alpha = 1e-8  # only used if alpha is not adapted
        alpha = self._compute_alpha(X, Y)

        super(WSABIL, self).__init__(base_gp=base_gp, X=X, Y=Y, lower_bound=alpha)

    def _compute_alpha(self, X: np.ndarray, Y: np.ndarray) -> float:
        """Compute the offset :math:`\alpha`.

        :param X: Observation locations, shape (n_points, input_dim)
        :param Y: Values of observations, shape (n_points, 1)
        :return: The offset :math:`\alpha`.
        """
        if self.adapt_alpha:
            return 0.8 * min(Y)[0]
        return self._small_alpha

    def compute_warping_params(self, X: np.ndarray, Y: np.ndarray) -> dict:
        r"""Computes the new :math:`\alpha` parameter from data.

        :param X: Observation locations, shape (n_points, input_dim).
        :param Y: Integrand observations at X, shape (n_points, 1).
        :return: Dictionary containing new value of :math:`\alpha`.
        """
        new_offset = self._compute_alpha(X, Y)
        return {"offset": new_offset}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================

import sys

from setuptools import find_packages, setup

from emukit.__version__ import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements/requirements.txt", "r") as req:
    requires = req.read().split("\n")

# enforce >Python3 for all versions of pip/setuptools
assert sys.version_info >= (3,), "This package requires Python 3."

setup(
    name="emukit",
    version=__version__,
    description="Toolkit for decision making under uncertainty.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/emukit/emukit",
    packages=find_packages(exclude=["test*"]),
    include_package_data=True,
    install_requires=requires,
    extras_require={"benchmarking": ["matplotlib"]},
    python_requires=">=3",
    license="Apache License 2.0",
    classifiers=(
        # https://pypi.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ),
)



import collections
import math
import time
from typing import Callable, Dict, List, Tuple

import GPy
import matplotlib.pyplot as plt
import numpy as np
import numpy.linalg as la
import scipy as sp
from GPy.util.univariate_Gaussian import cdfNormal, derivLogCdfNormal, logCdfNormal, std_norm_cdf, std_norm_pdf
from scipy.stats import multivariate_normal, mvn, norm

from .. import ComparisonGP, ComparisonGPEmukitWrapper, util
from .acquisition_function import AcquisitionFunction


def dK_dX(self, X: np.ndarray, X2: np.ndarray, dimX: int):
    """
    Derivative of RBF kernel with respect to input X

    :param X: First input
    :param X2: Second input
    :param dimX: dimension of X the derivative is wanted at
    :return: Derivative of the kernel with respect to the wanted dimension of X
    """
    r = self._scaled_dist(X, X2)
    K = self.K_of_r(r)
    dist = X[:, None, dimX] - X2[None, :, dimX]
    lengthscale2inv = (np.ones((X.shape[1])) / (self.lengthscale**2))[dimX]
    return -1.0 * K * dist * lengthscale2inv


GPy.kern.RBF.dK_dX = dK_dX

# Some helper functions
phi = lambda x: norm.cdf(x)  # cdfNormal(x)
npdf = lambda x, m, v: 1.0 / np.sqrt(2 * np.pi * v) * np.exp(-((x - m) ** 2) / (2 * v))


class ExpectationAcquisition(AcquisitionFunction):
    """
    Implements the general acquisition function that requires sampling from the posterior and uses reparametrization trick to compute the expectation required by the acquisition function
    """

    def dgp_ds_via_L(self, s: np.ndarray, dl: np.ndarray) -> np.ndarray:
        """
        Partial derivatives of the gp posterior samples with respect to the input locations given gradient with respect to cholesky the posterior covariance matrix.
        The derivative is computed using the generalised chain rule, meaning that derivative with respect to the posterior covariance matrix already exists

        :param s: Samples of observations from the posterior distribution of the model
        :param dl: Derivatives of the acquisition with respect to the covariance matrix
        :return: the derivatives of the gp samples with respect to the inputs
        """
        dls = np.zeros((s.shape[0], dl.shape[1], dl.shape[3], dl.shape[4]))
        for i in range(dl.shape[3]):
            for j in range(dl.shape[4]):
                dls[:, :, i, j] = np.matmul(dl[:, :, :, i, j], s[:, :, None])[:, :, 0]
        return dls

    def dgp_dL_via_Sigma(self, L: np.ndarray, L_inv: np.ndarray, dsigma: np.ndarray) -> np.ndarray:
        """
        Partial derivatives of the gp posterior samples with respect to the cholesky of the posterior covariance matrix given the partial derivative values with respect to the posterior covariance matrix.

        :param s: Samples of observations from the posterior distribution of the model
        :param L: Cholesky decomposition(s) of the posterior covariance matrix (samples)
        :param L_inv: Inverse(s) of Cholesky decomposition(s) of the posterior covariance matrix (samples)
        :param dsigma: Partial derivatives with respect to the posterior covariance matrix
        :return: the derivative of the gp samples with respect to the choleskies
        """
        E = np.tril(2 * np.ones((L.shape[1], L.shape[2])), -1) + np.eye(L.shape[2])
        dl = np.empty(dsigma.shape)  # np.empty((N,b,b,b,d))
        for i in range(dsigma.shape[3]):
            for j in range(dsigma.shape[4]):
                tmp1 = np.matmul(L_inv, dsigma[:, :, :, i, j])  # N x b x b
                tmp2 = np.matmul(tmp1, np.swapaxes(L_inv, 1, 2))  # N x b x b
                tmp3 = tmp2 * E[None, :, :]  # N x b x b
                dl[:, :, :, i, j] = 0.5 * np.matmul(L, tmp3)  # N x b x b
        return dl  # N x b x b

    def dgp_dSigma(self, x: np.ndarray, X: np.ndarray, kern: GPy.kern.Kern, w_inv: np.ndarray) -> np.ndarray:
        """
        Partial derivatives of the gp posterior samples with respect to the posterior covariance matrix

        :param x: The locations the samples are taken at
        :param X: The locations used to train the GP model
        :param kern: Prior covariance matrix
        :param w_inv: inverses of the woodbury matrix of the model
        :return: the derivative of the gp samples with respect to the matrix
        """
        N, b, d, n = w_inv.shape[0], x.shape[0], x.shape[1], X.shape[0]
        dkxX_dx = np.empty((b, n, d))
        dkxx_dx = np.empty((b, b, d))
        for i in range(d):
            dkxX_dx[:, :, i] = kern.dK_dX(x, X, i)
            dkxx_dx[:, :, i] = kern.dK_dX(x, x, i)
        K = kern.K(x, X)

        grad = np.empty((N, b, d))
        dsigma = np.zeros((N, b, b, b, d))
        for i in range(b):
            for j in range(d):
                Ks = np.zeros((b, n))
                Ks[i, :] = dkxX_dx[i, :, j]
                dKss_dxi = np.zeros((b, b))
                dKss_dxi[i, :] = dkxx_dx[i, :, j]
                dKss_dxi[:, i] = dkxx_dx[i, :, j].T
                dKss_dxi[i, i] = 0
                dsigma[:, :, :, i, j] = (
                    dKss_dxi[None, :, :]
                    - np.matmul(np.matmul(Ks[None, :, :], w_inv), (K.T)[None, :, :])
                    - np.matmul(np.matmul(K[None, :, :], w_inv), (Ks.T)[None, :, :])
                )
        return dsigma

    def dgp_dmean(self, kern: GPy.kern.Kern, w_vec: np.ndarray, x: np.ndarray, X: np.ndarray) -> np.ndarray:
        """
        Partial derivatives of the gp posterior samples with respect to the posterior mean

        :param kern: Prior covariance matrix
        :param w_vec: woodbury vectors of the posterior of the model
        :param x: The locations the samples are taken at
        :param X: The locations used to train the GP model
        :return: the derivative of the gp samples with respect to the mean
        """
        N, b, d, n = w_vec.shape[0], x.shape[0], x.shape[1], X.shape[0]
        dkxX_dx = np.empty((b, n, d))
        dmu = np.zeros((N, b, b, d))
        for i in range(d):
            dkxX_dx[:, :, i] = kern.dK_dX(x, X, i)
            for j in range(b):
                dmu[:, j, j, i] = np.matmul(dkxX_dx[j, :, i][None, :], w_vec[:, :, None]).flatten()  # d
        return dmu  # N x b x b x d

    def dgp_dx(
        self,
        s: np.ndarray,
        L: np.ndarray,
        L_inv: np.ndarray,
        w_vec: np.ndarray,
        w_inv: np.ndarray,
        kern: GPy.kern.Kern,
        x: np.ndarray,
        X: np.ndarray,
    ) -> np.ndarray:
        """
        Partial derivatives of the acquisition function values with respect to the sampled observations from the GP using generalised chain rule.

        :param s: Samples of observation from the posterior distribution of the model
        :param L: Cholesky decomposition(s) of the posterior covariance matrix (samples)
        :param L_inv: Inverse(s) of Cholesky decomposition(s) of the posterior covariance matrix (samples)
        :param w_vec: woodbury vectors of the posterior of the model
        :param w_inv: inverses of the woodbury matrix of the model
        :param kern: Prior covariance matrix
        :param x: The locations the samples are taken at
        :param X: The locations used to train the GP model
        :return: the derivative of the acquisition function values with respect to the samples
        """
        dmu_dx = self.dgp_dmean(kern, w_vec, x, X)
        dsigma_dx = self.dgp_dSigma(x, X, kern, w_inv)  # N x b x b x b x d
        dl_dx = self.dgp_dL_via_Sigma(L, L_inv, dsigma_dx)  # N x b x b x b x d
        dls_dx = self.dgp_ds_via_L(s, dl_dx)  # N x b x b x d
        return dmu_dx + dls_dx  # N x b x b x d

    def acquisition_fun(self, x: np.ndarray, m: ComparisonGP) -> Tuple[float, np.ndarray]:
        """
        Computes the acquisition function value and acquisition function derivatives at the input locations

        :param x: input location the acquisition and its derivative are evaluated at
        :param m: the GP model which posterior is used
        :return: Tuple containing the acquisition function value and its gradients
        """
        pools = self.pool_size
        N = self.acq_samples
        if m.name != "MCMC":
            pools = 1
        else:
            if pools != -1:
                N = N // pools
                Np = 1
            else:
                pools = 1
                Np = N

        x = np.array(x)
        # Assign values as class variables so there is no need to pass them
        b, d, n = (
            x.shape[0],
            x.shape[1],
            m.X.shape[0],
        )  # size of the batch, dimensions of x, number of samples in the training data

        opt_val = 0
        grad = np.zeros((b, d))
        for i in range(pools):
            # Take model specific variables
            if m.name == "MCMC":
                # We take predictive mu and L (Sigma = L L^T) of the posterior
                mu, L, w_inv, w_vec = m.posterior._get_mu_L(x, N=Np, woodbury_inv=True)
                L_inv = np.empty(L.shape)
                for i in range(L.shape[0]):
                    L_inv[i, :, :] = la.inv(L[i, :, :])
            else:
                mu, Sigma = m.predict_noiseless(x, full_cov=True)
                if len(Sigma.shape) == 3:
                    Sigma = Sigma[:, :, 0]
                L = la.cholesky(Sigma)
                L_inv = la.inv(L)
                mu, L, L_inv = mu[None, :, 0], L[None, :, :], L_inv[None, :, :]
                w_inv = m.posterior.woodbury_inv[None, :, :]
                w_vec = m.posterior.woodbury_vector[None, :, 0]
                if len(w_inv.shape) > 3:
                    w_inv = w_inv[:, :, :, 0]

            # Sample from N(0,I)
            s = np.random.normal(size=(N, b))
            s2 = np.random.normal(size=(N, b))
            # Generate samples for stochastic estimation of the expectation from the
            f = mu + np.matmul(L, s[:, :, None])[:, :, 0]
            y = f + np.sqrt(m.sigma2s[0]) * s2[:, :]

            sqrt_nu = np.sqrt(m.sigma2s[0] * np.ones(b))

            # Compute acquisition function value
            opt_val += self.opt_val(f, y, m, sqrt_nu)

            grad += self.opt_val_grad(s, f, y, m, sqrt_nu, L, L_inv, w_vec, w_inv, x)

        return -opt_val / pools, -grad / pools

    def acquisition_fun_flat(self, X: np.ndarray, m: ComparisonGP) -> Tuple[float, np.ndarray]:
        """
        Wrapper for acquisition_fun, where X is considered as a vector

        :param X: input location the acquisition and its derivative are evaluated at
        :param m: the GP model which posterior is used
        :return: Tuple containing the acquisition function value and its gradients
        """
        X = np.atleast_2d(X)
        n = m.X.shape[1]
        k = X.shape[1] // n
        opt_val = []
        dpdX = np.empty((0, n * k))
        for i in range(X.shape[0]):
            X_i = X[i, :]
            start = time.time()
            (opt_val_i, dpdX_i) = self.acquisition_fun(X_i.reshape(k, n), m)
            opt_val += [opt_val_i]
            dpdX = np.concatenate((dpdX, dpdX_i.flatten().reshape((1, -1))), axis=0)
        return np.array(opt_val).reshape((-1, 1)), dpdX

    def acq_fun_optimizer(
        self, m: ComparisonGP, bounds: np.ndarray, batch_size: int, get_logger: Callable
    ) -> np.ndarray:
        """
        Implements the optimization scheme for the sampling based acquisition function

        :param m: The model which posterior is used by the acquisition function (from which the samples are drawn from)
        :param bounds: the optimization bounds of the new sample
        :param batch_size: How many points are there in the batch
        :param get_logger: Function for receiving the legger where the prints are forwarded.
        :return: optimized locations
        """

        X = None  # Will hold the final choice
        y = None  # Will hold the expected improvement of the final choice

        # Run local gradient-descent optimizer multiple times
        # to avoid getting stuck in a poor local optimum
        # Tile bounds to match batch size
        # print(self.batch_size)

        bounds_tiled = np.tile(bounds, (batch_size, 1))

        xss = []
        for j in range(self.acq_opt_restarts):
            # Initial point of the optimization
            X0 = util.random_sample(bounds, batch_size)
            get_logger().info("acq opt restart {}".format(j))
            try:
                f_, _ = self.acquisition_fun_flat(X0.flatten(), m)
                get_logger().debug("Starting from: {} with acq {}".format(X0.flatten(), f_))
                res = sp.optimize.minimize(
                    fun=self.acquisition_fun_flat,
                    x0=X0.flatten(),
                    args=(m),
                    method="L-BFGS-B",
                    jac=True,
                    bounds=bounds_tiled,
                    options=self.optimizer_options,
                )
                X0 = res.x.reshape(batch_size, len(bounds))
                y0 = res.fun
                get_logger().debug("Ended to: {} with acq {}".format(X0, y0))
                # Update X if the current local minimum is
                # the best one found so far
                X0 = X0.reshape(batch_size, len(bounds))
                if X is None or y0 < y:
                    X = X0
                    y = y0
            except Exception as e:
                get_logger().error("Solver failed. Below the exception\n{}".format(e))
        # Assert that at least one optimization run succesfully
        assert X is not None
        return X

    def evaluate(self, x: np.ndarray, model: ComparisonGPEmukitWrapper) -> np.ndarray:
        """
        Computes the acquisition function at a batch.
        :param x: points where the acquisition is evaluated.
        :param batch_size: How many points are there in the batch.
        :return: acquisition function value evaluated at the batch
        """
        return self.acquisition_fun_flat(x, model.model)[0]

    def evaluate_with_gradients(self, x: np.ndarray, model: ComparisonGPEmukitWrapper) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes the acquisition function at a batch.
        :param x: points where the acquisition is evaluated.
        :param batch_size: How many points are there in the batch.
        :return: A tuple containing acquisition function value evaluated at the batch and its gradient
        """
        return self.acquisition_fun_flat(x, model.model)


class SumOfVariances(ExpectationAcquisition):
    """
    This class implements the Sum of Variances acquisition function
    """

    def opt_val(self, f: np.ndarray, y: np.ndarray, m: ComparisonGP, sqrt_nu: np.ndarray) -> float:
        """
        computes the acquisition function value by sampling

        :param f: latent function samples of the posterior
        :param y: observations from the latent functio (latent function values corrupted with noise)
        :param m: the model which posterior is used
        :param sqrt_nu: noise std for the observations
        :return: Sum of variances at the input
        """
        liks = np.zeros(f.shape[1])
        for i in range(f.shape[1]):  # Loop through each member being the batch winner (lower value better)
            mi = [k for k in range(f.shape[1]) if k is not i]  # select all indices but the batch winner
            prs = np.prod(
                phi((f[:, mi] - y[:, i, None]) / sqrt_nu[None, mi]), axis=1
            )  # Equation 3 in the draft (or the simplified version of it at the end of Section 2.2)
            prs2 = prs**2
            liks[i] = (
                np.mean(prs2, axis=0) - np.mean(prs, axis=0) ** 2
            )  # Variance of p(y_i \leq y_j \forall i \neq j) weighted by p(y_i \leq y_j \forall i \neq j)
        return np.sum(liks)

    def opt_val_grad(
        self,
        s: np.ndarray,
        f: np.ndarray,
        y: np.ndarray,
        m: ComparisonGP,
        sqrt_nu: np.ndarray,
        L: np.ndarray,
        L_inv: np.ndarray,
        w_vec: np.ndarray,
        w_inv: np.ndarray,
        x: np.ndarray,
    ) -> np.ndarray:
        """
        computes the acquisition function value by sampling

        :param f: latent function samples of the posterior
        :param y: observations from the latent functio (latent function values corrupted with noise)
        :param m: the model which posterior is used
        :param sqrt_nu: noise std for the observations
        :param L: Cholesky decomposition(s) of the posterior covariance matrix (samples)
        :param L_inv: Inverse(s) of Cholesky decomposition(s) of the posterior covariance matrix (samples)
        :param w_vec: woodbury vectors of the posterior of the model
        :param w_inv: inverses of the woodbury matrix of the model
        :param x: The locations the samples are taken at
        :return: Gradient of Sum of variances at the input
        """

        N, b, d, n = f.shape[0], x.shape[0], x.shape[1], m.X.shape[0]

        # compute the likelihood of each batch sample (equation at the end of 2.2 in the draft)
        val_s = self.opt_val_s(f, y, m, sqrt_nu)

        # compute the gradient of the likelihood of each batch sample using generalized chain rule
        val_s_grad = self.opt_val_s_grad(s, f, y, m, sqrt_nu, L, L_inv, w_vec, w_inv, x)

        grad = np.zeros((b, d))
        for i in range(b):  # Loop through winner index and compute the gradient of the likelihood
            grad = (
                grad
                + np.mean(2.0 * val_s[:, i, None, None] * val_s_grad[:, i, :, :], axis=0)
                - 2.0 * np.mean(val_s[:, i], axis=0) * np.mean(val_s_grad[:, i, :, :], axis=0)
            )
        return grad

    def opt_val_s(self, f: np.ndarray, y: np.ndarray, m: ComparisonGP, sqrt_nu: np.ndarray) -> np.ndarray:
        """
        Computes $ \prod_{i=1, i \neq j}^q p(y_j < y_i)$ for each sample from the latent function.

        :param f: latent function samples of the posterior
        :param y: observations from the latent functio (latent function values corrupted with noise)
        :param m: the model which posterior is used
        :param sqrt_nu: noise std for the observations
        :return: product of probabilities of comparisons for each sample
        """
        opt_vals = np.zeros(f.shape)
        for i in range(f.shape[1]):
            mi = [k for k in range(f.shape[1]) if k is not i]
            opt_vals[:, i] = np.prod(phi((f[:, mi] - y[:, i, None]) / sqrt_nu[None, mi]), axis=1)
        return opt_vals

    def opt_val_s_grad(
        self,
        s: np.ndarray,
        f: np.ndarray,
        y: np.ndarray,
        m: ComparisonGP,
        sqrt_nu: np.ndarray,
        L: np.ndarray,
        L_inv: np.ndarray,
        w_vec: np.ndarray,
        w_inv: np.ndarray,
        x: np.ndarray,
    ) -> np.ndarray:
        """
        Computes the gradient of $\prod_{i=1, i \neq j}^q p(y_j < y_i)$ for each sample from the latent function.

        :param f: latent function samples of the posterior
        :param y: observations from the latent functio (latent function values corrupted with noise)
        :param m: the model which posterior is used
        :param sqrt_nu: noise std for the observations
        :return: gradient of product of probabilities of comparisons for each sample
        """
        N, b, d, n = f.shape[0], x.shape[0], x.shape[1], m.X.shape[0]
        # Compute gradients using the generalized chain rule:
        opt_val_s_gradient_f = self.dopt_val_s_df(f, y, m, sqrt_nu)  # N x b x b
        gradient_f_x = self.dgp_dx(s, L, L_inv, w_vec, w_inv, m.kern, x, m.X)  # N x b x b x d
        grad = np.zeros((N, b, b, d))
        for k in range(b):  # batch winner
            for i in range(b):  # batch member
                for j in range(d):  # dimension of the batch member
                    grad[:, k, i, j] = np.matmul(
                        opt_val_s_gradient_f[:, k, None, :], gradient_f_x[:, :, None, i, j]
                    ).flatten()
        return grad

    def dopt_val_s_df(self, f: np.ndarray, y: np.ndarray, m: ComparisonGP, sqrt_nu: np.ndarray) -> np.ndarray:
        """
        Computes the gradient of $\prod_{i=1, i \neq j}^q p(y_j < y_i)$ for each sample from the latent function with respect to the GP

        :param f: latent function samples of the posterior
        :param y: observations from the latent functio (latent function values corrupted with noise)
        :param m: the model which posterior is used
        :param sqrt_nu: noise std for the observations
        :return: gradient of product of probabilities of comparisons for each sample with respect to the GP
        """
        dl = np.zeros((f.shape[0], f.shape[1], f.shape[1]))  # samples, winner , derivative index
        for i in range(f.shape[1]):  # loop through winner index
            for j in range(f.shape[1]):  # loop through f to derivate with respect to
                if j is not i:
                    mij = [k for k in range(f.shape[1]) if (k is not i) and (k is not j)]
                    dl[:, i, i] += (
                        -1.0
                        / sqrt_nu[j]
                        * npdf((f[:, j] - y[:, i]) / sqrt_nu[j], 0, 1)
                        * np.prod(phi((f[:, mij] - y[:, i, None]) / sqrt_nu[None, mij]), axis=1)
                    )
                    dl[:, i, j] = (
                        1.0
                        / sqrt_nu[j]
                        * npdf((f[:, j] - y[:, i]) / sqrt_nu[j], 0, 1)
                        * np.prod(phi((f[:, mij] - y[:, i, None]) / sqrt_nu[None, mij]), axis=1)
                    )
        return dl  # n x b x b


class QExpectedImprovement(ExpectationAcquisition):
    """
    This class implements the q Expected Improvement acquisition function
    """

    def opt_val(self, f: np.ndarray, y: np.ndarray, m: ComparisonGP, sqrt_nu: np.ndarray) -> float:
        """
        computes the acquisition function value by sampling

        :param f: latent function samples of the posterior
        :param y: observations from the latent functio (latent function values corrupted with noise)
        :param m: the model which posterior is used
        :param sqrt_nu: noise std for the observations
        :return: q Expected Improvement at the input
        """
        yw = m.get_current_best()
        liks = np.zeros(f.shape)
        for i in range(f.shape[1]):  # Loop through batch winners (lower value better)
            mi = [k for k in range(f.shape[1]) if k is not i]
            liks[:, i] = (
                (y[:, i] < yw)
                * np.array(yw - y[:, i])
                * np.prod(phi((f[:, mi] - y[:, i, None]) / sqrt_nu[None, mi]), axis=1)
            )
        return np.mean(np.sum(liks, axis=1), axis=0)

    def opt_val_grad(
        self,
        s: np.ndarray,
        f: np.ndarray,
        y: np.ndarray,
        m: ComparisonGP,
        sqrt_nu: np.ndarray,
        L: np.ndarray,
        L_inv: np.ndarray,
        w_vec: np.ndarray,
        w_inv: np.ndarray,
        x: np.ndarray,
    ) -> np.ndarray:
        """
        computes the acquisition function value by sampling

        :param f: latent function samples of the posterior
        :param y: observations from the latent functio (latent function values corrupted with noise)
        :param m: the model which posterior is used
        :param sqrt_nu: noise std for the observations
        :param L: Cholesky decomposition(s) of the posterior covariance matrix (samples)
        :param L_inv: Inverse(s) of Cholesky decomposition(s) of the posterior covariance matrix (samples)
        :param w_vec: woodbury vectors of the posterior of the model
        :param w_inv: inverses of the woodbury matrix of the model
        :param x: The locations the samples are taken at
        :return: Gradient of q Expected Improvement at the input
        """
        N, b, d, n = f.shape[0], x.shape[0], x.shape[1], m.X.shape[0]
        # Compute gradients using the generalized chain rule:
        opt_val_gradient_f = self.dopt_val_df(f, y, m, sqrt_nu)  # N x b
        gradient_f_x = self.dgp_dx(s, L, L_inv, w_vec, w_inv, m.kern, x, m.X)  # N x b x b x d
        grad = np.zeros((N, b, d))
        for i in range(b):
            for j in range(d):
                grad[:, i, j] = np.matmul(opt_val_gradient_f[:, None, :], gradient_f_x[:, :, None, i, j]).flatten()
        return np.mean(grad, axis=0)

    def dopt_val_df(self, f: np.ndarray, y: np.ndarray, m: ComparisonGP, sqrt_nu: np.ndarray) -> np.ndarray:
        """
        computes the gradient of the acquisition function value with respect to each input location and gp sample

        :param f: latent function samples of the posterior
        :param y: observations from the latent functio (latent function values corrupted with noise)
        :param m: the model which posterior is used
        :param sqrt_nu: noise std for the observations
        :return: gradient of the acquisition function value with respect to each input location and gp sample
        """
        yw = m.get_current_best()
        dl = np.zeros(f.shape)
        for i in range(f.shape[1]):  # loop through f to derivate with respect to
            mi = [k for k in range(f.shape[1]) if (k is not i)]
            dl[:, i] = dl[:, i] - np.prod(phi((f[:, mi] - y[:, i, None]) / sqrt_nu[None, mi]), axis=1) * (y[:, i] < yw)
            for j in range(f.shape[1]):
                if j is not i:
                    mij = [k for k in range(f.shape[1]) if (k is not i) and (k is not j)]
                    dl[:, i] = (
                        dl[:, i]
                        - (y[:, i] < yw)
                        * (yw - y[:, i])
                        / sqrt_nu[j]
                        * npdf((f[:, j] - y[:, i]) / sqrt_nu[j], 0, 1)
                        * np.prod(phi((f[:, mij] - y[:, i, None]) / sqrt_nu[None, mij]), axis=1)
                        + (y[:, j] < yw)
                        * (yw - y[:, j])
                        / sqrt_nu[i]
                        * npdf((f[:, i] - y[:, j]) / sqrt_nu[i], 0, 1)
                        * np.prod(phi((f[:, mij] - y[:, j, None]) / sqrt_nu[None, mij]), axis=1)
                    )  # first line: i is winner second line: j is winner
        return dl  # n x b



"""GPy wrappers for the quadrature package."""

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import warnings
from typing import List, Optional, Tuple, Union

import GPy
import numpy as np
from scipy.linalg import lapack

from ..quadrature.interfaces import (
    IRBF,
    IBaseGaussianProcess,
    IBrownian,
    IProductBrownian,
    IProductMatern12,
    IProductMatern32,
    IProductMatern52,
)
from ..quadrature.kernels import (
    QuadratureBrownianLebesgueMeasure,
    QuadratureKernel,
    QuadratureProductBrownian,
    QuadratureProductBrownianLebesgueMeasure,
    QuadratureProductMatern12LebesgueMeasure,
    QuadratureProductMatern32LebesgueMeasure,
    QuadratureProductMatern52LebesgueMeasure,
    QuadratureRBFGaussianMeasure,
    QuadratureRBFLebesgueMeasure,
)
from ..quadrature.measures import BoxDomain, GaussianMeasure, IntegrationMeasure, LebesgueMeasure
from ..quadrature.typing import BoundsType


class BaseGaussianProcessGPy(IBaseGaussianProcess):
    """Wrapper for GPy's :class:`GPRegression` as required for some EmuKit quadrature methods.

    An instance of this class can be passed as :attr:`base_gp` to a :class:`WarpedBayesianQuadratureModel` object.

    .. note::
        GPy's :class:`GPRegression` cannot take ``None`` as initial values for X and Y. Thus, we initialize
        them with some arbitrary values. These will be re-set in the :class:`WarpedBayesianQuadratureModel`.

    :param kern: An EmuKit quadrature kernel.
    :param gpy_model: A GPy GP regression model.
    :param noise_free: If ``False``, the observation noise variance will be treated as a model parameter,
                       if ``True`` the noise is set to 1e-10, defaults to ``True``.
    """

    def __init__(self, kern: QuadratureKernel, gpy_model: GPy.models.GPRegression, noise_free: bool = True):
        super().__init__(kern=kern)
        if noise_free:
            gpy_model.Gaussian_noise.constrain_fixed(1.0e-10)
        self.gpy_model = gpy_model

        if isinstance(kern, QuadratureProductBrownian):
            if kern.offset != 0:
                raise ValueError(
                    "The wrapper BaseGaussianProcessGPy does not support EmuKit product Brownian "
                    "motion kernels with non-zero offset as these are not supported in GPy."
                )

    @property
    def X(self) -> np.ndarray:
        """The data nodes."""
        return self.gpy_model.X

    @property
    def Y(self) -> np.ndarray:
        """The data evaluations at the nodes."""
        return self.gpy_model.Y

    @property
    def observation_noise_variance(self) -> float:
        return self.gpy_model.Gaussian_noise[0]

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """Sets training data in model.

        :param X: New training features, shape (num_points, input_dim).
        :param Y: New training outputs, shape (num_points, 1).
        """
        self.gpy_model.set_XY(X, Y)

    def predict(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Predictive mean and covariance at the locations X_pred.

        :param X_pred: Points at which to predict, with shape (number of points, input_dim).
        :return: Predictive mean, predictive variances shapes (num_points, 1) and (num_points, 1).
        """
        return self.gpy_model.predict(X_pred, full_cov=False)

    def predict_with_full_covariance(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Predictive mean and covariance at the locations X_pred.

        :param X_pred: Points at which to predict, with shape (num_points, input_dim).
        :return: Predictive mean, predictive full covariance shapes (num_points, 1) and (num_points, num_points).
        """
        return self.gpy_model.predict(X_pred, full_cov=True)

    def solve_linear(self, z: np.ndarray) -> np.ndarray:
        lower_chol = self.gpy_model.posterior.woodbury_chol
        return lapack.dtrtrs(lower_chol.T, (lapack.dtrtrs(lower_chol, z, lower=1)[0]), lower=0)[0]

    def graminv_residual(self) -> np.ndarray:
        return self.gpy_model.posterior.woodbury_vector

    def optimize(self) -> None:
        """Optimize the hyperparameters of the GP."""
        self.gpy_model.optimize()


class RBFGPy(IRBF):
    r"""Wrapper of the GPy RBF kernel as required for some EmuKit quadrature methods.

    .. math::
        k(x, x') = \sigma^2 e^{-\frac{1}{2}\sum_{i=1}^{d}r_i^2},

    where :math:`d` is the input dimensionality,
    :math:`r_i = \frac{x_i-x_i'}{\lambda_i}` is the scaled vector difference of dimension :math:`i`,
    :math:`\lambda_i` is the :math:`i` th element of the ``lengthscales`` property
    and :math:`\sigma^2` is the ``variance`` property.

    :param gpy_rbf: An RBF kernel from GPy.
    """

    def __init__(self, gpy_rbf: GPy.kern.RBF):
        self.gpy_rbf = gpy_rbf

    @property
    def lengthscales(self) -> np.ndarray:
        if self.gpy_rbf.ARD:
            return self.gpy_rbf.lengthscale.values
        return np.full((self.gpy_rbf.input_dim,), self.gpy_rbf.lengthscale[0])

    @property
    def variance(self) -> float:
        return self.gpy_rbf.variance.values[0]

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        return self.gpy_rbf.K(x1, x2)


class ProductMatern12GPy(IProductMatern12):
    r"""Wrapper of the GPy Exponential (a.k.a Matern12) product kernel as required for some
    EmuKit quadrature methods.

    The product kernel is of the form
    :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = e^{-r_i}.

    :math:`d` is the input dimensionality,
    :math:`r_i:=\frac{|x_i - x'_i|}{\lambda_i}`,
    :math:`\sigma^2` is the ``variance`` property and :math:`\lambda_i` is the :math:`i` th element
    of the ``lengthscales`` property.

    :param gpy_matern: An Exponential (a.k.a. Matern12) product kernel from GPy. For :math:`d=1` this is equivalent
                       to an Exponential kernel. For :math:`d>1`, this is *not* a :math:`d`-dimensional
                       Exponential kernel but a product of :math:`d` 1-dimensional Exponential kernels with differing
                       active dimensions constructed as k1 * k2 * ... .
                       Make sure to unlink all variances except the variance of the first kernel k1 in the product
                       as the variance of k1 will be used to represent :math:`\sigma^2`. If you are unsure what
                       to do, use the :attr:`lengthscales` and :attr:`variance` parameter instead.
                       If :attr:`gpy_matern` is not given, the :attr:`lengthscales` argument is used.
    :param lengthscales: If :attr:`gpy_matern` is not given, a product Matern12 kernel will be constructed with
                       the given lengthscales. The number of elements need to be equal to the dimensionality
                       :math:`d`. If :attr:`gpy_matern` is given, this input is disregarded.
    :param variance: The variance of the product kernel. Only used if :attr:`gpy_matern` is not given. Defaults to 1.
    """

    def __init__(
        self,
        gpy_matern: Optional[Union[GPy.kern.Exponential, GPy.kern.Prod]] = None,
        lengthscales: Optional[np.ndarray] = None,
        variance: Optional[float] = None,
    ):
        if gpy_matern is None and lengthscales is None:
            raise ValueError("Either lengthscales or a GPy product matern kernel must be given.")

        # product kernel from parameters
        if gpy_matern is None:

            input_dim = len(lengthscales)
            if input_dim < 1:
                raise ValueError("'lengthscales' must contain at least 1 value.")

            # default variance
            if variance is None:
                variance = 1.0

            gpy_matern = GPy.kern.Exponential(
                input_dim=1, active_dims=[0], lengthscale=lengthscales[0], variance=variance
            )
            for dim in range(1, input_dim):
                k = GPy.kern.Exponential(input_dim=1, active_dims=[dim], lengthscale=lengthscales[dim])
                k.unlink_parameter(k.variance)
                gpy_matern = gpy_matern * k

        self.gpy_matern = gpy_matern

    @property
    def lengthscales(self) -> np.ndarray:
        if isinstance(self.gpy_matern, GPy.kern.Exponential):
            return np.array([self.gpy_matern.lengthscale[0]])

        lengthscales = []
        for kern in self.gpy_matern.parameters:
            lengthscales.append(kern.lengthscale[0])
        return np.array(lengthscales)

    @property
    def variance(self) -> float:
        if isinstance(self.gpy_matern, GPy.kern.Exponential):
            return self.gpy_matern.variance[0]

        return self.gpy_matern.parameters[0].variance[0]

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        return self.gpy_matern.K(x1, x2)

    def _K_from_prod(self, x1: np.ndarray, x2: np.ndarray, skip: List[int] = None) -> np.ndarray:
        """The kernel k(x1, x2) evaluated at x1 and x2 computed as product from the
        individual 1d kernels.

        :param x1: First argument of the kernel, shape (n_points N, input_dim)
        :param x2: Second argument of the kernel, shape (n_points M, input_dim)
        :param skip: Skip these dimensions if specified.
        :returns: Kernel evaluated at x1, x2, shape (N, M).
        """
        if skip is None:
            skip = []
        K = np.ones([x1.shape[0], x2.shape[0]])
        for dim, kern in enumerate(self.gpy_matern.parameters):
            if dim in skip:
                continue
            K *= kern.K(x1, x2)

        # correct for missing variance
        if 0 in skip:
            K *= self.variance
        return K

    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        if isinstance(self.gpy_matern, GPy.kern.Exponential):
            return self._dK_dx1_1d(x1[:, 0], x2[:, 0], self.gpy_matern.lengthscale[0])[None, :, :]

        # product kernel
        dK_dx1 = np.ones([x1.shape[1], x1.shape[0], x2.shape[0]])
        for dim, kern in enumerate(self.gpy_matern.parameters):
            prod_term = self._K_from_prod(x1, x2, skip=[dim])  # N x M
            grad_term = self._dK_dx1_1d(x1[:, dim], x2[:, dim], kern.lengthscale[0])  # N x M
            dK_dx1[dim, :, :] *= prod_term * grad_term
        return dK_dx1


class ProductMatern32GPy(IProductMatern32):
    r"""Wrapper of the GPy Matern32 product kernel as required for some EmuKit quadrature methods.

    The product kernel is of the form
    :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = (1 + \sqrt{3}r_i ) e^{-\sqrt{3} r_i}.

    :math:`d` is the input dimensionality,
    :math:`r_i:=\frac{|x_i - x'_i|}{\lambda_i}`,
    :math:`\sigma^2` is the ``variance`` property and :math:`\lambda_i` is the :math:`i` th element
    of the ``lengthscales`` property.

    :param gpy_matern: A Matern32 product kernel from GPy. For :math:`d=1` this is equivalent to a
                       Matern32 kernel. For :math:`d>1`, this is *not* a :math:`d`-dimensional
                       Matern32 kernel but a product of :math:`d` 1-dimensional Matern32 kernels with differing
                       active dimensions constructed as k1 * k2 * ... .
                       Make sure to unlink all variances except the variance of the first kernel k1 in the product
                       as the variance of k1 will be used to represent :math:`\sigma^2`. If you are unsure what
                       to do, use the :attr:`lengthscales` and :attr:`variance` parameter instead.
                       If :attr:`gpy_matern` is not given, the :attr:`lengthscales` argument is used.
    :param lengthscales: If :attr:`gpy_matern` is not given, a product Matern32 kernel will be constructed with
                       the given lengthscales. The number of elements need to be equal to the dimensionality
                       :math:`d`. If :attr:`gpy_matern` is given, this input is disregarded.
    :param variance: The variance of the product kernel. Only used if :attr:`gpy_matern` is not given. Defaults to 1.
    """

    def __init__(
        self,
        gpy_matern: Optional[Union[GPy.kern.Matern32, GPy.kern.Prod]] = None,
        lengthscales: Optional[np.ndarray] = None,
        variance: Optional[float] = None,
    ):
        if gpy_matern is None and lengthscales is None:
            raise ValueError("Either lengthscales or a GPy product matern kernel must be given.")

        # product kernel from parameters
        if gpy_matern is None:

            input_dim = len(lengthscales)
            if input_dim < 1:
                raise ValueError("'lengthscales' must contain at least 1 value.")

            # default variance
            if variance is None:
                variance = 1.0

            gpy_matern = GPy.kern.Matern32(input_dim=1, active_dims=[0], lengthscale=lengthscales[0], variance=variance)
            for dim in range(1, input_dim):
                k = GPy.kern.Matern32(input_dim=1, active_dims=[dim], lengthscale=lengthscales[dim])
                k.unlink_parameter(k.variance)
                gpy_matern = gpy_matern * k

        self.gpy_matern = gpy_matern

    @property
    def lengthscales(self) -> np.ndarray:
        if isinstance(self.gpy_matern, GPy.kern.Matern32):
            return np.array([self.gpy_matern.lengthscale[0]])

        lengthscales = []
        for kern in self.gpy_matern.parameters:
            lengthscales.append(kern.lengthscale[0])
        return np.array(lengthscales)

    @property
    def variance(self) -> float:
        if isinstance(self.gpy_matern, GPy.kern.Matern32):
            return self.gpy_matern.variance[0]

        return self.gpy_matern.parameters[0].variance[0]

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        return self.gpy_matern.K(x1, x2)

    def _K_from_prod(self, x1: np.ndarray, x2: np.ndarray, skip: List[int] = None) -> np.ndarray:
        """The kernel k(x1, x2) evaluated at x1 and x2 computed as product from the
        individual 1d kernels.

        :param x1: First argument of the kernel, shape (n_points N, input_dim)
        :param x2: Second argument of the kernel, shape (n_points M, input_dim)
        :param skip: Skip these dimensions if specified.
        :returns: Kernel evaluated at x1, x2, shape (N, M).
        """
        if skip is None:
            skip = []
        K = np.ones([x1.shape[0], x2.shape[0]])
        for dim, kern in enumerate(self.gpy_matern.parameters):
            if dim in skip:
                continue
            K *= kern.K(x1, x2)

        # correct for missing variance
        if 0 in skip:
            K *= self.variance
        return K

    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        if isinstance(self.gpy_matern, GPy.kern.Matern32):
            return self._dK_dx1_1d(x1[:, 0], x2[:, 0], self.gpy_matern.lengthscale[0])[None, :, :]

        # product kernel
        dK_dx1 = np.ones([x1.shape[1], x1.shape[0], x2.shape[0]])
        for dim, kern in enumerate(self.gpy_matern.parameters):
            prod_term = self._K_from_prod(x1, x2, skip=[dim])  # N x M
            grad_term = self._dK_dx1_1d(x1[:, dim], x2[:, dim], kern.lengthscale[0])  # N x M
            dK_dx1[dim, :, :] *= prod_term * grad_term
        return dK_dx1


class ProductMatern52GPy(IProductMatern52):
    r"""Wrapper of the GPy Matern52 product kernel as required for some EmuKit quadrature methods.

    The product kernel is of the form
    :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = (1 + \sqrt{5} r_i + \frac{5}{3} r_i^2) \exp(- \sqrt{5} r_i).

    :math:`d` is the input dimensionality,
    :math:`r_i:=\frac{|x_i - x'_i|}{\lambda_i}`,
    :math:`\sigma^2` is the ``variance`` property and :math:`\lambda_i` is the :math:`i` th element
    of the ``lengthscales`` property.

    :param gpy_matern: A Matern52 product kernel from GPy. For :math:`d=1` this is equivalent to a
                       Matern52 kernel. For :math:`d>1`, this is *not* a :math:`d`-dimensional
                       Matern52 kernel but a product of :math:`d` 1-dimensional Matern52 kernels with differing
                       active dimensions constructed as k1 * k2 * ... .
                       Make sure to unlink all variances except the variance of the first kernel k1 in the product
                       as the variance of k1 will be used to represent :math:`\sigma^2`. If you are unsure what
                       to do, use the :attr:`lengthscales` and :attr:`variance` parameter instead.
                       If :attr:`gpy_matern` is not given, the :attr:`lengthscales` argument is used.
    :param lengthscales: If :attr:`gpy_matern` is not given, a product Matern52 kernel will be constructed with
                       the given lengthscales. The number of elements need to be equal to the dimensionality
                       :math:`d`. If :attr:`gpy_matern` is given, this input is disregarded.
    :param variance: The variance of the product kernel. Only used if :attr:`gpy_matern` is not given. Defaults to 1.
    """

    def __init__(
        self,
        gpy_matern: Optional[Union[GPy.kern.Matern52, GPy.kern.Prod]] = None,
        lengthscales: Optional[np.ndarray] = None,
        variance: Optional[float] = None,
    ):
        if gpy_matern is None and lengthscales is None:
            raise ValueError("Either lengthscales or a GPy product matern kernel must be given.")

        # product kernel from parameters
        if gpy_matern is None:

            input_dim = len(lengthscales)
            if input_dim < 1:
                raise ValueError("'lengthscales' must contain at least 1 value.")

            # default variance
            if variance is None:
                variance = 1.0

            gpy_matern = GPy.kern.Matern52(input_dim=1, active_dims=[0], lengthscale=lengthscales[0], variance=variance)
            for dim in range(1, input_dim):
                k = GPy.kern.Matern52(input_dim=1, active_dims=[dim], lengthscale=lengthscales[dim])
                k.unlink_parameter(k.variance)
                gpy_matern = gpy_matern * k

        self.gpy_matern = gpy_matern

    @property
    def lengthscales(self) -> np.ndarray:
        if isinstance(self.gpy_matern, GPy.kern.Matern52):
            return np.array([self.gpy_matern.lengthscale[0]])

        lengthscales = []
        for kern in self.gpy_matern.parameters:
            lengthscales.append(kern.lengthscale[0])
        return np.array(lengthscales)

    @property
    def variance(self) -> float:
        if isinstance(self.gpy_matern, GPy.kern.Matern52):
            return self.gpy_matern.variance[0]

        return self.gpy_matern.parameters[0].variance[0]

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        return self.gpy_matern.K(x1, x2)

    def _K_from_prod(self, x1: np.ndarray, x2: np.ndarray, skip: List[int] = None) -> np.ndarray:
        """The kernel k(x1, x2) evaluated at x1 and x2 computed as product from the
        individual 1d kernels.

        :param x1: First argument of the kernel, shape (n_points N, input_dim)
        :param x2: Second argument of the kernel, shape (n_points M, input_dim)
        :param skip: Skip these dimensions if specified.
        :returns: Kernel evaluated at x1, x2, shape (N, M).
        """
        if skip is None:
            skip = []
        K = np.ones([x1.shape[0], x2.shape[0]])
        for dim, kern in enumerate(self.gpy_matern.parameters):
            if dim in skip:
                continue
            K *= kern.K(x1, x2)

        # correct for missing variance
        if 0 in skip:
            K *= self.variance
        return K

    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        if isinstance(self.gpy_matern, GPy.kern.Matern52):
            return self._dK_dx1_1d(x1[:, 0], x2[:, 0], self.gpy_matern.lengthscale[0])[None, :, :]

        # product kernel
        dK_dx1 = np.ones([x1.shape[1], x1.shape[0], x2.shape[0]])
        for dim, kern in enumerate(self.gpy_matern.parameters):
            prod_term = self._K_from_prod(x1, x2, skip=[dim])  # N x M
            grad_term = self._dK_dx1_1d(x1[:, dim], x2[:, dim], kern.lengthscale[0])  # N x M
            dK_dx1[dim, :, :] *= prod_term * grad_term
        return dK_dx1


class BrownianGPy(IBrownian):
    r"""Wrapper of the GPy Brownian motion kernel as required for some EmuKit quadrature methods.

    .. math::
        k(x, x') = \sigma^2 \operatorname{min}(x, x')\quad\text{with}\quad x, x' \geq 0,

    where :math:`\sigma^2` is the ``variance`` property.

    :param gpy_brownian: A Brownian motion kernel from GPy.
    """

    def __init__(self, gpy_brownian: GPy.kern.Brownian):
        self.gpy_brownian = gpy_brownian

    @property
    def variance(self) -> float:
        return self.gpy_brownian.variance[0]

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        return self.gpy_brownian.K(x1, x2)


class ProductBrownianGPy(IProductBrownian):
    r"""Wrapper of the GPy Brownian product kernel as required for some EmuKit quadrature methods.

    The product kernel is of the form
    :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = \operatorname{min}(x_i-c, x_i'-c)\quad\text{with}\quad x_i, x_i' \geq c,

    :math:`d` is the input dimensionality,
    :math:`\sigma^2` is the ``variance`` property
    and :math:`c` is the ``offset`` property.

    :param gpy_brownian: A Brownian product kernel from GPy. For :math:`d=1` this is equivalent to a
                         Brownian kernel. For :math:`d>1`, this is a product of :math:`d` 1-dimensional Brownian
                         kernels with differing active dimensions constructed as k1 * k2 * ... .
                         Make sure to unlink all variances except the variance of the first kernel k1 in the product
                         as the variance of k1 will be used to represent :math:`\sigma^2`. If you are unsure what
                         to do, use the :attr:`input_dim` and :attr:`variance` parameter instead.
                         If :attr:`gpy_brownian` is not given, the :attr:`variance` and :attr:`input_dim`
                         argument is used.
    :param offset: The offset :math:`c` of the kernel. Defaults to 0.
    :param variance: The variance of the product kernel. Only used if :attr:`gpy_brownian` is not given. Defaults to 1.
    :param input_dim: The input dimension. Only used if :attr:`gpy_brownian` is not given.
    """

    def __init__(
        self,
        gpy_brownian: Optional[Union[GPy.kern.Brownian, GPy.kern.Prod]] = None,
        offset: float = 0.0,
        variance: Optional[float] = None,
        input_dim: Optional[int] = None,
    ):

        if gpy_brownian is not None:
            if input_dim is not None or variance is not None:
                warnings.warn("gpy_brownian and variance is given. The variance will be ignore.")
        else:
            if input_dim is None or variance is None:
                raise ValueError(
                    "Please provide a GPy product Brownian kernel or alternitvely the the variance and input_dim."
                )

        # default variance
        if variance is None:
            variance = 1.0

        # product kernel from parameters
        if gpy_brownian is None:
            gpy_brownian = GPy.kern.Brownian(input_dim=1, active_dims=[0], variance=variance)
            for dim in range(1, input_dim):
                k = GPy.kern.Brownian(input_dim=1, active_dims=[dim])
                k.unlink_parameter(k.variance)
                gpy_brownian = gpy_brownian * k

        self.gpy_brownian = gpy_brownian
        self._offset = offset

    @property
    def variance(self) -> float:
        if isinstance(self.gpy_brownian, GPy.kern.Brownian):
            return self.gpy_brownian.variance[0]

        return self.gpy_brownian.parameters[0].variance[0]

    @property
    def offset(self) -> float:
        return self._offset

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        return self.gpy_brownian.K(x1 - self._offset, x2 - self.offset)

    def _K_from_prod(self, x1: np.ndarray, x2: np.ndarray, skip: List[int] = None) -> np.ndarray:
        """The kernel k(x1, x2) with offset=0 evaluated at x1 and x2 computed as product from the
        individual 1d kernels.

        :param x1: First argument of the kernel, shape (n_points N, input_dim)
        :param x2: Second argument of the kernel, shape (n_points M, input_dim)
        :param skip: Skip these dimensions if specified.
        :returns: Kernel evaluated at x1, x2, shape (N, M).
        """
        if skip is None:
            skip = []
        K = np.ones([x1.shape[0], x2.shape[0]])
        for dim, kern in enumerate(self.gpy_brownian.parameters):
            if dim in skip:
                continue
            K *= kern.K(x1, x2)

        # correct for missing variance
        if 0 in skip:
            K *= self.variance
        return K

    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        if isinstance(self.gpy_brownian, GPy.kern.Brownian):
            return self._dK_dx1_1d(x1[:, 0], x2[:, 0])[None, :, :]

        # product kernel
        x1 = x1 - self.offset
        x2 = x2 - self.offset
        dK_dx1 = np.ones([x1.shape[1], x1.shape[0], x2.shape[0]])
        for dim, kern in enumerate(self.gpy_brownian.parameters):
            prod_term = self._K_from_prod(x1, x2, skip=[dim])  # N x M
            grad_term = self._dK_dx1_1d(x1[:, dim], x2[:, dim])  # N x M
            dK_dx1[dim, :, :] *= prod_term * grad_term
        return dK_dx1

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        """The gradient of the diagonal of the kernel (the variance) v(x):=k(x, x) evaluated at x.

        :param x: The locations where the gradient is evaluated, shape (n_points, input_dim).
        :return: The gradient of the diagonal of the kernel evaluated at x, shape (input_dim, n_points).
        """
        if isinstance(self.gpy_brownian, GPy.kern.Brownian):
            return self.variance * np.ones((x.shape[1], x.shape[0]))

        x = x - self.offset
        dKdiag_dx = np.ones((x.shape[1], x.shape[0]))
        for dim, kern in enumerate(self.gpy_brownian.parameters):
            prod_term = np.prod(x, axis=1) / x[:, dim]  # N,
            grad_term = 1.0
            dKdiag_dx[dim, :] *= prod_term * grad_term
        return self.variance * dKdiag_dx


# === convenience functions start here


def create_emukit_model_from_gpy_model(
    gpy_model: GPy.models.GPRegression,
    integral_bounds: Optional[BoundsType] = None,
    measure: Optional[IntegrationMeasure] = None,
    integral_name: str = "",
) -> BaseGaussianProcessGPy:
    """Wraps a GPy model and returns an EmuKit quadrature model.

    :param gpy_model: A GPy Gaussian process regression model ``GPy.models.GPRegression``.
    :param integral_bounds: List of d tuples, where d is the dimensionality of the integral and the tuples contain the
                            lower and upper bounds of the integral
                            i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_d, ub_d)].
                            Only used if ``measure`` is not given in which case the unnormalized Lebesgue measure is used.
    :param measure: An integration measure. Either ``measure`` or ``integral_bounds`` must be given.
                    If both ``integral_bounds`` and ``measure`` are given, ``integral_bounds`` is disregarded.
    :param integral_name: The (variable) name(s) of the integral.
    :return: An EmuKit GP model for quadrature with GPy backend.

    """

    if (integral_bounds is None) and (measure is None):
        raise ValueError("Either measure or integral bounds must be given.")

    if (integral_bounds is not None) and (measure is not None):
        warnings.warn("Both measure and integral bounds are given. Bounds are being ignored.")

    if measure is None:
        domain = BoxDomain(name="", bounds=integral_bounds)
        measure = LebesgueMeasure(domain=domain)

    def _check_is_gpy_product_kernel(k, k_type):
        is_type = isinstance(gpy_model.kern, k_type)
        if isinstance(k, GPy.kern.Prod):
            all_type = all(isinstance(kern, k_type) for kern in k.parameters)
            all_univariante = all(kern.input_dim == 1 for kern in k.parameters)
            if all_type and all_univariante:
                is_type = True
        return is_type

    # wrap standard kernel
    # RBF
    qkern_emukit = None
    if isinstance(gpy_model.kern, GPy.kern.RBF):
        skern_emukit = RBFGPy(gpy_model.kern)
        if isinstance(measure, LebesgueMeasure):
            qkern_emukit = QuadratureRBFLebesgueMeasure(skern_emukit, measure, integral_name)
        elif isinstance(measure, GaussianMeasure):
            qkern_emukit = QuadratureRBFGaussianMeasure(skern_emukit, measure, integral_name)

    # Univariate Matern12 or ProductMatern12
    elif _check_is_gpy_product_kernel(gpy_model.kern, GPy.kern.Exponential):
        skern_emukit = ProductMatern12GPy(gpy_model.kern)
        if isinstance(measure, LebesgueMeasure):
            qkern_emukit = QuadratureProductMatern12LebesgueMeasure(skern_emukit, measure, integral_name)

    # Univariate Matern32 or ProductMatern32
    elif _check_is_gpy_product_kernel(gpy_model.kern, GPy.kern.Matern32):
        skern_emukit = ProductMatern32GPy(gpy_model.kern)
        if isinstance(measure, LebesgueMeasure):
            qkern_emukit = QuadratureProductMatern32LebesgueMeasure(skern_emukit, measure, integral_name)

    # Univariate Matern52 or ProductMatern52
    elif _check_is_gpy_product_kernel(gpy_model.kern, GPy.kern.Matern52):
        skern_emukit = ProductMatern52GPy(gpy_model.kern)
        if isinstance(measure, LebesgueMeasure):
            qkern_emukit = QuadratureProductMatern52LebesgueMeasure(skern_emukit, measure, integral_name)

    # Brownian
    elif isinstance(gpy_model.kern, GPy.kern.Brownian):
        skern_emukit = BrownianGPy(gpy_model.kern)
        if isinstance(measure, LebesgueMeasure):
            qkern_emukit = QuadratureBrownianLebesgueMeasure(skern_emukit, measure, integral_name)

    # ProductBrownian
    elif _check_is_gpy_product_kernel(gpy_model.kern, GPy.kern.Brownian):
        skern_emukit = ProductBrownianGPy(gpy_model.kern)
        if isinstance(measure, LebesgueMeasure):
            qkern_emukit = QuadratureProductBrownianLebesgueMeasure(skern_emukit, measure, integral_name)

    else:
        raise ValueError(f"There is no GPy wrapper for the provided GPy kernel ({gpy_model.kern.name}).")

    if qkern_emukit is None:
        raise ValueError(f"Kernel embedding not available for provided kernel-measure combination.")

    # wrap the base-gp model
    return BaseGaussianProcessGPy(kern=qkern_emukit, gpy_model=gpy_model)



# Copyright 2020 Opsani, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Any, List, Optional, Tuple, Union

import numpy as np

from .categorical_parameter import CategoricalParameter
from .continuous_parameter import ContinuousParameter
from .discrete_parameter import DiscreteParameter
from .encodings import OneHotEncoding
from .parameter import Parameter


class BanditParameter(Parameter):
    """
    A multivariate parameter consisting of a restricted domain of the full Cartesian product of its
    constituent sub-parameters
    """

    def __init__(self, name: str, domain: np.ndarray, sub_parameter_names: Optional[List[str]] = None):
        """
        :param name: Name of parameter
        :param domain: List of tuples representing valid values
        :param sub_parameter_names: List of parameters, must correspond to domain if provided,
            otherwise will be reflected from the domain
        """
        super().__init__(name)
        if not isinstance(domain, np.ndarray):
            raise ValueError("Domain must be a 2D np.ndarray, got type: {}".format(type(domain)))
        if not domain.ndim == 2:
            raise ValueError("Domain must be a 2D np.ndarray, got dimensions: {}".format(domain.shape))
        self.domain = domain  # each column is homogeneously typed thanks to numpy.ndarray
        self.parameters = self._create_parameters(domain, sub_parameter_names)
        self._sub_parameter_names = sub_parameter_names

    def __str__(self):
        msg = f"<BanditParameter: {self.name} ndim={self.domain.ndim}"
        if self._sub_parameter_names:
            msg = msg + f" ({','.join(self._sub_parameter_names)})"
        msg = msg + ">"
        return msg

    def __repr__(self):
        return f"BanditParameter({self.name}, {self.domain}, {self._sub_parameter_names})"

    def _create_parameter_names(self, domain: np.ndarray) -> List[str]:
        """
        Create names for sub-parameters, used when names are not already provided.

        :param domain: 2D array (n by p) of valid states, each row represents one valid state
        :returns: List of names
        """
        return [f"{self.name}_{i}" for i in range(domain.shape[1])]

    def _create_parameters(self, domain: np.ndarray, parameter_names: Optional[List[str]]) -> List[Parameter]:
        """
        Reflect parameters from domain.

        :param domain: 2D array (n by p) of valid states, each row represents one valid state
        :param parameter_names: Optional list of names for sub-parameters. If not provided,
            sub-parameter names will be automatically generated.
        :returns: List of sub-parameters
        """
        parameters = []
        parameter_names = parameter_names if parameter_names else self._create_parameter_names(domain)
        if domain.shape[1] != len(parameter_names):
            raise ValueError(
                "Provided domain shape {} != number of parameter names {}".format(domain.shape[1], len(parameter_names))
            )
        for cix, parameter_name in enumerate(parameter_names):
            sub_param_domain = domain[:, cix]
            domain_unq = np.unique(sub_param_domain)
            if np.issubdtype(sub_param_domain.dtype, np.number):  # make discrete
                parameter = DiscreteParameter(name=parameter_name, domain=domain_unq)
            else:  # make categorical
                encoding = OneHotEncoding(domain_unq)
                parameter = CategoricalParameter(name=parameter_name, encoding=encoding)
                raise NotImplementedError("Categorical sub-parameters not yet fully supported")
                # NOTE Categorical sub-parameters not yet implemented because inputs are
                # homogeneously typed np.ndarrays rather than structured arrays. In the future,
                # using structured arrays for all inputs may be more appropriate.
            parameters.append(parameter)
        return parameters

    @property
    def model_parameters(self) -> List:
        return self.parameters

    def check_in_domain(self, x: Union[np.ndarray, float, list]) -> Union[bool, np.ndarray]:
        """
        Checks if all the points in x lie in the domain set

        :param x:    1d numpy array of points to check
                  or 2d numpy array with shape (n_points, 1) of points to check
                  or float of single point to check
        :return: A boolean value which indicates whether all points lie in the domain
        """
        if isinstance(x, np.ndarray):
            if x.ndim == 2 and x.shape[1] == 1:
                x = x.ravel()
            elif x.ndim == 2 and x.shape[1] > 1:
                result = np.array([self.check_in_domain(xx) for xx in x])
                return result.any()
            elif x.ndim > 1:
                raise ValueError("Expected x shape (n,) or (n, 1), actual is {}".format(x.shape))
            if x.shape[0] != self.domain.shape[1]:
                raise ValueError(
                    "Received x with dimension {}, expected dimension is {}".format(x.shape[0], self.domain.shape[1])
                )
            result = (self.domain == x).all(axis=1).any()
        elif isinstance(x, float):
            if self.domain.shape[1] > 1:
                raise ValueError(
                    "Received x with dimension 1, expected dimension is {}".format(x.shape[0], self.domain.shape[1])
                )
            result = (self.domain == x).all(axis=1).any()
        elif isinstance(x, list):
            result = (self.domain == x).all(axis=1).any()
        else:
            raise ValueError("Unsupported type for point x: {}".format(type(x)))
        return result

    @property
    def bounds(self) -> List[Tuple]:
        """
        Calculate the limiting bounds of the sub-parameters

        :returns: a list containing tuples (min, max) for each constituent sub-parameter
        """
        return [pb for p in self.parameters for pb in p.bounds]

    def round(self, x: np.ndarray) -> np.ndarray:
        """
        Rounds each row in x to represent a valid value for this bandit variable. Note that this
        valid value may be 'far' from the suggested value.

        :param x: A 2d array NxD to be rounded (D is len(self.parameters))
        :returns: An array NxD where each row represents a value from the domain
                  that is closest to the corresponding row in x
        """
        if x.ndim != 2:
            raise ValueError("Expected 2d array, got {}".format(x.ndim))

        if x.shape[1] != self.dimension:
            raise ValueError("Expected {} column array, got {}".format(self.dimension, x.shape[1]))

        x_rounded = []
        for row in x:
            dists = np.sqrt(np.sum((self.domain - row) ** 2))
            rounded_value = min(self.domain, key=lambda d: np.linalg.norm(d - row))
            x_rounded.append(rounded_value)

        if not all([self.check_in_domain(xr) for xr in x_rounded]):
            raise ValueError("Rounding error encountered, not all rounded values in domain.")
        return np.row_stack(x_rounded)

    @property
    def dimension(self) -> int:
        """
        Calculate the aggregate dimensionality of the sub-parameters

        :returns: dimensionality of the BanditParameter
        """
        d = 0
        for p in self.parameters:
            if isinstance(p, ContinuousParameter):
                d += 1
            elif isinstance(p, DiscreteParameter):
                d += 1
            elif isinstance(p, CategoricalParameter):
                d += p.dimension
            else:
                raise Exception("Parameter type {type(p)} not supported.")
        return d

    def sample_uniform(self, point_count: int) -> np.ndarray:
        """
        Generates multiple uniformly distributed random parameter points.

        :param point_count: number of data points to generate.
        :returns: Generated points with shape (point_count, num_features)
        """
        return self.domain[np.random.choice(self.domain.shape[0], point_count)]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, Union

import numpy as np

from ...core.acquisition import Acquisition
from ...core.interfaces import IDifferentiable, IModel


class ModelVariance(Acquisition):
    """
    This acquisition selects the point in the domain where the predictive variance is the highest
    """

    def __init__(self, model: Union[IModel, IDifferentiable]) -> None:
        self.model = model

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        _, variance = self.model.predict(x)
        return variance

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple:
        _, variance = self.model.predict(x)
        _, dvariance_dx = self.model.get_prediction_gradients(x)
        return variance, dvariance_dx

    @property
    def has_gradients(self):
        return isinstance(self.model, IDifferentiable)
